"""
/NasaAmes1001/read.py
$Id: read.py 2557 2020-12-02 23:54:45Z pe $

Basic NASA Ames 1001 file reader class.
Handles low level file I/O.

History:
V.1.0.0  2013-03-05  pe  initial version

"""

import re
import math
import sys
import datetime
from fileformats.AsciiRead import AsciiRead, BackposError
from nilutility.ndecimal import Decimal, InvalidOperation
from nilutility.datatypes import HexInt
from nilutility.logging_helper import INFO_MINUS
from nilutility.msg_condenser import MessageRecord, MessageCondenser
from .base import NasaAmes1001Error, NasaAmes1001Base


MSG_NUM_VAL_ERR = 0
MSG_CVT_ERR = 1
MSG_CVT2_ERR = 2
MSG_XINC_ERR = 3
MSG_XINCDX_ERR = 4
MSG_VMISS_ERR = 5
MSG_VFORMAT_ERR = 6


class NasaAmesMessageRecord(MessageRecord):
    """
    Derived MessageRecord class for EbasIO:
    Adds column number to keys and line number to data.
    """
    KEYS = ['msgfunc', 'msg_id', 'cnum']
    DATA = ['message', 'line']

class NasaAmesMessageCondenser(MessageCondenser):
    """
    Message condenser for NasaAmes logging
    """

    def deliver_msg(self, record, condensed):
        """
        Deliver a single message.
        Parameters:
            record     the message record
            condensed  bool, condensed message
        Returns:
            None
        """
        record.msgfunc(record.message, lnum_override=record.line)
        if condensed:
            lst = self.dict[record.key]
            record.msgfunc(
                "The previous message repeats {} times{}. "
                "First occurrences in lines: {}, ..."
                .format(
                    len(lst)-1,
                    " for DATA[{}]".format(record.cnum) \
                        if record.cnum is not None else "",
                    ", ".join(
                        [str(x.line) for x in lst[0:self.threshold]])),
                lnum_override=record.line, counter_increase=len(lst)-1)
            # use same lnum as above message, otherwise the lines get parted
            # in submit-tool


class NasaAmes1001Read(NasaAmes1001Base):  # pylint: disable=R0902
    # R0902: Too many instance attributes
    """
    Partial class NasaAmes1001: Part for reading NasaAmes1001 files into object.
    """

    def __init__(self, *args, **kwargs):
        """
        Paramaters:
            *args, **kwargs, passd through to base class
        Returns:
            None
        """
        super(NasaAmes1001Read, self).__init__(*args, **kwargs)
        self.nbsp_warned = False
        self.msg_condenser = None
        self.vmiss_vals = None

    def read(self, filespec, encoding=None, skip_data=False,
             condense_messages=0):
        """
        Reads the file according to NasaAmes 1001 standard.
        Parameters:
            filespec    file name (incl. path) or file like object (stream)
            encoding    force a specific encoding to be used
            skip_data   skip reading of data (speedup, if data are not needed at
                        application level). First and last line are read.
        Returns:
            None
        Raises:
            IOError           file open error
            NasaAmesError     on any unrecoverable read error
        """
        # reset the object for reading
        self.init_state()
        self.init_data()

        self.encoding = encoding
        self.enc_line = None

        self.msg_condenser = NasaAmesMessageCondenser(
            threshold=condense_messages, record_cls=NasaAmesMessageRecord)

        with AsciiRead(
            filespec, encodings=('ascii', 'utf-8', 'utf-16', 'utf-32'),
            fallback_encoding='iso8859-15', strict_enc=True,
            ignore_bom=2, cvt_nbsp=2, strip=True, tabsize=1,
            error_callback=self.ascii_read_error_callback,
            warning_callback=self.ascii_read_warning_callback) as self.file:
            self._read(skip_data)

    def _read(self, skip_data):
        """
        Reads the file from opened file handle.
        This method is only needed to extract the reading functionality in order
        to wrap the 'with open(filename)' around it in the read() method.
        Parameters:
            skip_data   skip reading of data (speedup, if data are not needed at
                        application level). First and last line are read.
        Returns:
            None
        Raises:
            NasaAmesError     on any unrecoverable read error
        """
        for self.state in range(len(self.__class__.SECTIONS)):
            statename = self.curr_state_name()
            try:
                method = getattr(self, '_read_' + statename.lower())
            except AttributeError:
                # If no dedicated _read method exists: store one line as string\
                # Those are: ONAME, ORG, SNAME, MNAME, XNAME
                self.data[statename] = self._read_header_line()
                self.logger.debug(u'%s=%s', statename, self.data[statename])
            else:
                if statename != 'DATA' or not skip_data:
                    method()
                else:
                    method(skip_data)
        # check dx:
        if self.data.DATA != []  and len(self.data.DATA[0]) > 1 and \
           self.data.DX != None:
            # dx is only valid, if number of samples is greater 1
            # also make sure there was no syntax error in DX (!=None)
            dx_ = self.dx_from_data()
            if not skip_data and not self._dx_equal(self.data.DX, dx_):
                if self.data.DX == 0 and \
                   not self.strictness & self.__class__.STRICT_DX:
                    # if DX=0 and not strict: warning and set correctly
                    self.warning("DX ({}) ".format(self.data.DX) + \
                        "specified in line 8 is inconsistent with data"
                        " - setting {}".format(dx_))
                    self.data.DX = dx_
                else:
                    # if DX!=0 or strictness is requested: error
                    # (a set DX != 0 must always be obeyed!)
                    self.error("DX ({}) ".format(self.data.DX) + \
                        "specified in line 8 is inconsistent with data"
                        " - should be {}".format(dx_))
        if self.errors:
            raise NasaAmes1001Error("Exiting because of previous errors")

    def _read_nlhead(self):
        """
        Read the first line from a NASA Ames 1001 file: Check for 1001 specifier
        and store NLHEAD.
        Parameters:
            None
        Returns:
            None
        Raises:
            NasaAmes1001Error   on syntax error
        """
        try:
            line = self._readline()
        except EOFError:
            self.error("unexpected end of file", exception=True)
        reg = re.match(r'^(\d+) +1001$', line)
        if not reg:
            self.error("this is not a NASA Ames 1001 file", exception=True)
        self.data.NLHEAD = int(reg.group(1))
        if self.data.NLHEAD < 15:
            self.warning('NLHEAD < 15 seems too little, expect '+\
                '"incomplete NASA Ames header" error to follow')
        self.logger.debug('NLHEAD=%s', self.data.NLHEAD)

    def _read_ivol_nvol(self):
        """
        Read the IVOL/NVOL line from a NASA Ames 1001 file.
        Parameters:
            None
        Returns:
            None
        """
        reg = re.match(r'^(\d+) +(\d+)$', self._read_header_line())
        if not reg:
            self.error("IVOL/NVOL syntax error")
        else:
            self.data.IVOL = int(reg.group(1))
            self.data.NVOL = int(reg.group(2))
        self.logger.debug('IVOL=%s, NVOL=%s',
                          # use %s instead of %d (else None fails)
                          self.data.IVOL, self.data.NVOL)

    def _read_date_rdate(self):
        """
        Read the DATE/RDATE line from a NASA Ames 1001 file: Check syntax and
        store DATE and RDATE as datetime.datetime objects.
        Parameters:
            None
        Returns:
            None
        """
        reg = re.match(
            r'^(\d\d\d\d) +(\d\d) +(\d\d) +(\d\d\d\d) +(\d\d) +(\d\d) *$',
            self._read_header_line())
        if not reg:
            self.error('DATE/RDATE syntax error')
            return # stop parsing of this line
        (dyyyy, dmo, ddd, ryyyy, rmo, rdd) = reg.groups()
        try:
            self.data.DATE = datetime.datetime(year=int(dyyyy), month=int(dmo),
                                               day=int(ddd))
        except ValueError as excpt:
            self.error('DATE error: ' + str(excpt))
        self.logger.debug('DATE=%s', self.data.DATE)
        try:
            self.data.RDATE = datetime.datetime(year=int(ryyyy), month=int(rmo),
                                                day=int(rdd))
        except ValueError as excpt:
            self.error('RDATE error: ' + str(excpt))
        self.logger.debug('RDATE=%s', self.data.RDATE)

    def _read_dx(self):
        """
        Read the DX line from a NASA Ames 1001 file.
        Parameters:
            None
        Returns:
            None
        Raises:
            NasaAmes1001Error   on syntax error
        """
        reg = re.match(r'^(\d*(\.\d*)?)$', self._read_header_line())
        if not reg:
            # this is a non recoverable error
            self.error('DX syntax error')
            return
        if not reg.group(1):
            self.error('DX may not be empty')
            return
        try:
            self.data.DX = float(reg.group(1))
        except ValueError:
            self.error('DX syntax error')
            return
        self.logger.debug('DX=%f', self.data.DX)

    def _read_nv(self):
        """
        Read the NV line from a NASA Ames 1001 file.
        Parameters:
            None
        Returns:
            None
        Raises:
            NasaAmes1001Error   on syntax error
        """
        reg = re.match(r'^(\d+)$', self._read_header_line())
        if not reg:
            # this is a non recoverable error
            self.error('NV syntax error', exception=True)
        self.data.NV = int(reg.group(1))
        self.logger.debug('NV=%d', self.data.NV)

    def _read_vscal(self):
        """
        Read the VSACL line from a NASA Ames 1001 file.
        Parameters:
            None
        Returns:
            None
        Raises:
            NasaAmes1001Error   on syntax error
        """
        self.data.VSCAL = []
        vscal = self._read_header_line().split()
        if len(vscal) != self.data.NV:
            self.error('number of VSCAL elements ({}) does not match NV ({})'.\
                        format(len(vscal), self.data.NV))
        for i in range(max(len(vscal), self.data.NV)):
            value = None
            if i < len(vscal):
                try:
                    value = float(vscal[i])
                except ValueError:
                    self.error('VSCAL[{}]: not a float value'.format(i))
            if i < self.data.NV:
                self.data.VSCAL.append(value)
        self.logger.debug('VSCAL=%s', self.data.VSCAL)

    def _read_vmiss(self):
        """
        Read the VMISS line from a NASA Ames 1001 file.
        Parameters:
            None
        Returns:
            None
        Raises:
            NasaAmes1001Error   on syntax error
        """
        self.data.VMISS = []
        self.data.VFORMAT = []
        self.data.VFORMAT2 = []
        self.data.VREPR = []   # representation: float or Decimal
        vmiss = self._read_header_line().split()
        if len(vmiss) != self.data.NV:
            self.error('number of VMISS elements ({}) does not match NV ({})'.\
                        format(len(vmiss), self.data.NV))
        for i in range(max(len(vmiss), self.data.NV)):
            (val, fmt, fmt2, rep) = (None, None, None, float)
            if i < len(vmiss):
                try:
                    (val, fmt, fmt2, rep) = self._get_vmiss_fmt(vmiss[i])
                except ValueError as excpt:
                    self.error("VMISS[{}]: {}".format(
                        i, str(excpt)))
            if i < self.data.NV:
                self.data.VMISS.append(val)
                self.data.VFORMAT.append(fmt)
                self.data.VFORMAT2.append(fmt2)
                self.data.VREPR.append(rep)
        self.logger.debug('VMISS=%s', self.data.VMISS)
        self.logger.debug('VFORMAT=%s', self.data.VFORMAT)

    @staticmethod
    def _get_vmiss_fmt(vmiss_str):
        """
        Analyzes the vmiss string and returns information on it.
        Parameters:
            vmiss_str    string containing one vmiss value
        Returns:
            tupel: (val, fmt, fmt2, rep)
                value of the vmiss value (str)
                format string for the data variable (str)
                short format string for the data variable, (without padding,
                    just number of digits after comma) (str)
                representation of the data variable (data type)
        """
        (val, fmt, fmt2, rep) = (None, None, None, float)
        reg = re.match('^-?0[xX]([0-9a-fA-F]+)$', vmiss_str.strip())
        if reg:
            # hex format
            try:
                _ = HexInt(vmiss_str)
            except ValueError:
                raise ValueError("'{}: illegal number format".format(vmiss_str))
            vk_ = len(reg.group(1))
            if vk_ > math.log(sys.float_info.epsilon)/math.log(16)*-1:
                raise ValueError(
                    "'{}': long hex (>13 digits) is not supported".format(
                        vmiss_str))
            else:
                val = vmiss_str.strip()
                fmt = '#0{}x'.format(vk_ + 2)
                fmt2 = '#0{}X'.format(vk_ + 2)
                # fmt2: must be upper case X, that way we can avoid one upper()
                # when checking the data format
                rep = HexInt
        else:
            try:
                _ = float(vmiss_str)
            except ValueError:
                raise ValueError("'{}': illegal number format".format(
                    vmiss_str))
            val = vmiss_str.strip()
            reg = re.match(r"^-?(\d*)(\.(\d*))?$", val)
            if reg:
                vk_ = len(reg.group(1))
                nk_ = len(reg.group(3)) if reg.group(3) else 0
                kk_ = len(reg.group(2)) if reg.group(2) else 0
                fmt = '{}.{}f'.format(vk_ + kk_ + 1, nk_)
                fmt2 = '.{}f'.format(nk_)
            else:
                reg = re.match(r"^-?(\d*)(\.(\d*))?((e|E)[+-]?(\d+))$", val)
                if reg:
                    vk_ = len(reg.group(1))
                    nk_ = len(reg.group(3)) if reg.group(3) else 0
                    kk_ = len(reg.group(2)) if reg.group(2) else 0
                    exp = len(reg.group(6)) if reg.group(6) else 2
                    if exp < 2:
                        exp = 2
                    fmt = '{}.{}E'.format(vk_ + kk_ + exp + 3, nk_)
                    fmt2 = '.{}E'.format(nk_)
                    # fmt2: must be upper case E, that way we can avoid one
                    # upper() when checking the data format
                else:
                    raise ValueError("'{}': illegal number format".format(val))
            if vk_ + nk_ > math.log10(sys.float_info.epsilon)*-1:
                # this variable can not be accurately expressed as
                # float
                rep = Decimal
        return (val, fmt, fmt2, rep)

    def _read_vname(self):
        """
        Reads the VNAME lines from a NASA Ames 1001 file.
        Parameters:
            None
        Returns:
            None
        Raises:
            NasaAmes1001Error   on syntax error
        """
        self.data.VNAME = []
        for i in range(self.data.NV):
            self.data.VNAME.append(self._read_header_line())
            if re.match(r"^\d$", self.data.VNAME[-1]):
                # this is a "fuzzy" error: this is mainly for cases where NV is
                # set too low, and NSCOM is interpreted as variable name
                # (which really messes up the whole parsing after this point)
                self.error(
                    "VNAME[{}]: variable name syntax error. '{}' is not a "
                    "valid VNAME (check if NV in line 10 is set correctly)".\
                        format(i, self.data.VNAME[-1]))
                raise NasaAmes1001Error("Exiting because of previous errors")
        self.logger.debug('VNAME=%s', self.data.VNAME)

    def _read_scoml(self):
        """
        Read the complete SCOML comment block.
        Parameters:
            None
        Returns:
            None
        """
        line = self._read_header_line()
        if not re.match(r'^\d+$', line):
            self.error('NSCOML: syntax error', exception=True)
        nscoml = int(line)
        self.data.SCOML = []
        for _ in range(nscoml):
            self.data.SCOML.append(self._read_header_line())
        self.logger.debug('SCOML=%s', self.data.SCOML)

    def _read_ncoml(self):
        """
        Read the complete NCOML comment block.
        Parameters:
            None
        Returns:
            None
        """
        line = self._read_header_line()
        if not re.match(r'^\d+$', line):
            self.error('NNCOML: syntax error', exception=True)
        nncoml = int(line)
        if self.data.NLHEAD != self.lnum + nncoml:
            self.error('NLHEAD and NNCOM are inconsistent', exception=True)
        self.data.NCOML = []
        for _ in range(nncoml):
            self.data.NCOML.append(self._read_header_line())
        self.logger.debug('NCOML=%s', self.data.NCOML)

    def _read_data(self, skip_data=False):
        """
        Read the complete data block.
        Parameters:
            skip_data   skip reading of data (speedup, if data are not needed at
                        application level). First and last line are read.
        Returns:
            None
        """
        self.data.DATA = []
        for _ in range(self.data.NV + 1):
            self.data.DATA.append([])
        skipped = False

        self.vmiss_vals = [_typeconvert(self.data.VMISS[i], self.data.VREPR[i])
                           if self.data.VMISS[i] is not None else None
                           for i in range(self.data.NV)]
        while True:
            if self.lnum % 1000 == 0:
                self.logger.log(INFO_MINUS, "reading line %d", self.lnum)
            try:
                line = self._readline()
            except EOFError:
                break
            self._add_data(line, skip_data)
            if skip_data and not skipped:
                try:
                    self.file.goto_line(-2, allow_backpos=False)
                except BackposError:
                    return
                skipped = True
        self.vmiss_vals = None
        # write all condensed messages and reset condenser object:
        self.msg_condenser.deliver()

    def _add_data(self, line, skip_data=False):
        """
        Add 1 line of data to internal store.
        Parameters:
            line     line of data from file (str)
            skip_data    only read first and last line (speedup, if data are not
                         needed on application level)
        Returns:
            None
        """
        def strip(val):
            """
            Strip a leading + sign, then substitute all 0 if not like "0.".
            This is needed for the number format check below.
            """
            # optimize runtime: quickly return in usual cases, and do not copy
            if val[0] not in ('+', '0') or val[0:2] == '0.':
                return val
            # else: not so performant...
            i = 0
            if val[0] == '+':
                i = 1
            while val[i] == '0' and len(val) > i+1:
                i += 1
            if i > 0 and val[i-1] == '0' and val[i] in ('.', 'x', 'X'):
                i -= 1
            return val[i:]

        if self.strictness & self.__class__.STRICT_NUMFORMAT:
            values = [strip(val) for val in line.split()]
        else:
            values = line.split()
        if len(values) != self.data.NV + 1:
            self.msg_condenser.add(
                self.error, MSG_NUM_VAL_ERR, None,
                'data line does not contain {} values'.format(self.data.NV + 1),
                self.lnum)
        if not values:
            return
        try:
            self._add_x_value(values[0], skip_data)
        except ValueError:
            return  # error parsing x-value: skip this data line
        for cnum in range(1, self.data.NV + 1):
            value = None
            if cnum < len(values):
                try:
                    value = _typeconvert(values[cnum], self.data.VREPR[cnum-1])
                except ValueError as excpt:
                    self.msg_condenser.add(
                        self.error, MSG_CVT_ERR, cnum,
                        "DATA[{}]: {}".format(cnum, str(excpt)),
                        self.lnum)

            if value is not None:
                if self.vmiss_vals[cnum-1] is not None and \
                   value == self.vmiss_vals[cnum-1]:
                    value = None
                # check dependent variable:
                elif self.vmiss_vals[cnum-1] is not None and \
                   value > self.vmiss_vals[cnum-1]:
                    self.msg_condenser.add(
                        self.error, MSG_VMISS_ERR, cnum,
                        "DATA[{}]: value ({}) > VMISS ({}, missing value "
                        "specified in line 12) is not allowed."
                        .format(cnum, value, self.data.VMISS[cnum-1]),
                        self.lnum)
                # number format checks:
                # convert the numeric value back to string using the format
                # specifier based on the VMISS. The strings should be identical
                # This is a very expansive operation, and it's performed for
                # each value!
                # 2018/07: PE: optimizations:
                #  - used the new VFORMAT2 instead of VFORMAT (which spared a
                #    strip operation)
                #  - used capital E (exponent) and X (HexInt) in VFORMAT2 and
                #    uses upper on the right hand comparison argument,
                #    So instead of .lower() != .lower, we need just one upper()
                #  - unfortunately found no way to avoid the expansove string
                #    operations
                elif (self.strictness & self.__class__.STRICT_NUMFORMAT) and \
                    self.data.VFORMAT2[cnum-1] and \
                    '{:{}}'.format(value, self.data.VFORMAT2[cnum-1]) != \
                        values[cnum].upper():
                    self.msg_condenser.add(
                        self.error, MSG_VFORMAT_ERR, cnum,
                        "DATA[{}]: value ({}) does not match VMISS format "
                        "({}, missing value specified in line 12)".format(
                            cnum, values[cnum], self.data.VMISS[cnum-1]),
                        self.lnum)
            self.DATA[cnum].append(value)

    def _add_x_value(self, valuestr, skip_data=False):
        """
        Adds next x value to the internal data structure.
        Parameters:
            valuestr    the value to be added (as string read from the file)
            skip_data   only read first and last line (speedup, if data are not
                        needed on application level)
        Returns:
            None
        Raises:
            ValueError   if string is malformatted (can not be interpreted as
                         float)
        """
        # generate xmindig while reading line by line
        try:
            mindig = minmaxdig(valuestr)[0]
        except ValueError:
            self.msg_condenser.add(
                self.error, MSG_CVT_ERR, 0,
                "DATA[0]: '{}' is not a float value. Skipping this "
                "data line".format(valuestr), self.lnum)
            raise
        if self.xmindig is None:
            self.xmindig = mindig
        else:
            self.xmindig = min(self.xmindig, mindig)

        value = None
        try:
            value = _typeconvert(valuestr, float)
        except ValueError as excpt:
            self.msg_condenser.add(
                self.error, MSG_CVT2_ERR, 0,
                'DATA[0]: ' + str(excpt) + ". Skipping this data line",
                self.lnum)
            raise

        if self.data.DATA[0] and self.data.DATA[0][-1] >= value:
            self.msg_condenser.add(
                self.error, MSG_XINC_ERR, 0,
                'DATA[0]: independent variable must be strictly increasing',
                self.lnum)
        if not skip_data and self.data.DX != None and self.data.DX != 0 and \
           self.data.DATA[0] and \
           not self._dx_equal(self.data.DX, abs(value-self.data.DATA[0][-1])):
            self.msg_condenser.add(
                self.error, MSG_XINCDX_ERR, 0,
                'DATA[0]: independent variable does not increase by specified '
                'DX', self.lnum)
        self.data.DATA[0].append(value)

    def _read_header_line(self):
        """
        Reads next header line from file.
        Raises Exceptions if number of header lines is not correct.
        Parameters:
            None
        Returns:
            header line (string)
            None if EOF
        Raises:
            NasaAmesError  if not enough lines in header (according to NLHEAD)
            EOFError       on unexpected EOF
        """
        if self.lnum >= self.data.NLHEAD:
            self.error("incomplete NASA Ames header (missing from " +\
                       "{}, NLHEAD was {})".format(self.curr_state_name(),
                                                   self.data.NLHEAD))
        try:
            return self._readline()
        except EOFError:
            self.error("unexpected end of file", exception=True)

    def _readline(self):
        """
        Reads one line from file and fills the objects line attribute.
        Parameters:
            None
        Returns:
            line read (str)
        Raises:
            EOFError
        """
        self.lnum += 1
        return self.file.readline()

    def ascii_read_error_callback(self, msg):
        """
        Callback method for AsciiRead error message handling.
        Parameters:
            msg    message
        Returns:
            False: prevent AsciiRead from handling the error by itself
        """
        self.error(msg, lnum=True, exception=True)
        return False

    def ascii_read_warning_callback(self, msg):
        """
        Callback method for AsciiRead warning message handling.
        Parameters:
            msg    message
        Returns:
            False: prevent AsciiRead from handling the warning by itself
        """
        self.warning(msg, lnum=True)
        return False

def _typeconvert(value, targettype):
    """
    Type conversion for data values. float, Decimal and HexInt are supported.
    Parameters:
        value        source value (str)
        targettype   target data type (float, Decimal, HexInt)
    """
    try:
        conv = targettype(value)
    except (ValueError, InvalidOperation):
        # float raises ValueError, Decimal raises InvalidOperation
        # we always raise ValueError:
        raise ValueError("'{}' is not a {} value"
                         .format(value, targettype.__name__))
    if math.isnan(conv) or math.isinf(conv):
        # Problem: different string representations of inf, -inf and nan are
        # possible in float and Decimal (e.g. Inf, -inf, infinity, NaN, nan).
        # Solution: check after conversion with math.isinf and math.isnan
        # instead of parsing the string representation.
        # math does support all used types: int (HexInt based on int), float
        # and Decimal.
        raise ValueError("'{}' is not allowed as {} value in NasaAmes1001".\
              format(value, targettype.__name__))
    return conv

def minmaxdig(floatstr):
    """
    Calculates the minimum and maximum digit for a float value represented in
    a string.
    Parameters:
        floatstr    string representing a float number
    Returns:
        magnitude of minimum and maximum digit used in the number
    Raises:
        ValueError   if string is malformatted (can not be interpreted as float)
    """
    reg = re.match(r'^-?([0-9]*)(\.([0-9]*))?((E|e)(([+-])?[0-9]*))?$',
                   floatstr)
    if not reg:
        raise ValueError('string is not a float representation')
    maxdig = len(reg.group(1)) if reg.group(1) else 0
    maxdig += int(reg.group(6)) if reg.group(6) else 0
    mindig = len(reg.group(3)) * -1 if reg.group(3) else 0
    mindig += int(reg.group(6)) if reg.group(6) else 0
    return (mindig, maxdig)



"""
$Id: base.py 2759 2021-11-30 11:41:16Z pe $
EANET base class
"""

from ..basefile.file import EbasFile
from ...base import EbasIOError
from ebas.domain.masterdata.org import EbasMasterOR
from ebas.domain.masterdata.st import EbasMasterST

class EbasEanetReadError(EbasIOError):
    """
    Exception class for EANET read errors.
    """
    pass



class EbasEanetBase(EbasFile):  # pylint: disable=W0223
    # W0223: Method ... is abstract in base class but is not overridden
    # This is also an abstract class, just does not raise NotImplementedError
    """
    Base class.
    """

    def _set_method(self, metadata):
        """
        Set the method from analytical lab code and method name
        """
        if 'method' not in metadata or not metadata.method:
            metadata.method = ''
            if '#ana_lab_code' in metadata and metadata['#ana_lab_code']:
                metadata.method += metadata['#ana_lab_code']
            elif 'lab_code' in metadata and metadata['lab_code']:
                self.warning(
                    "#ana_lab_code not set, using lab_code for method")
                metadata.method += metadata['lab_code']
            else:
                self.error(
                    "#ana_lab_code and lab_code not set")
            if '#method_name' in metadata and metadata['#method_name']:
                metadata.method += '_' + metadata['#method_name']
            else:
                self.error(
                    "#method_name not set")
        if '#ana_lab_code' in metadata:
            del metadata['#ana_lab_code']
        if '#method_name' in metadata:
            del metadata['#method_name']

    def _set_station(self, metadata):
        """
        Set all station metadata:
            - by station code from config from ebas masterdata
            - then override from config
        """
        
        if 'station_code' in metadata and metadata.station_code:
            try:
                master_st = EbasMasterST()[metadata.station_code]
            except KeyError:
                self.warning(
                    "Station '{}' not found in EBAS "
                    "masterdata.".format(metadata.station_code))
            else:
                # station_code already set, no need to copy
                # platform code is special, needs to be mapped manually:
                if 'platform_code' not in metadata:
                    metadata.platform_code = \
                        master_st.ST_STATION_CODE[:-1] + \
                        master_st.PF_PLATFORM
                # generic mapping:
                maps = [
                    ('station_name', 'ST_NAME'),
                    ('station_wdca_id', 'ST_WDCA_ID'),
                    ('station_gaw_id', 'ST_GAW_ID'),
                    ('station_gaw_name', 'ST_GAW_NAME'),
                    ('station_airs_id', 'ST_AIRS_ID'),
                    ('station_other_ids', 'ST_OTHER_ID'),
                    ('station_state_code', 'ST_STATE_CODE'),
                    ('station_landuse', 'SL_STATION_LANDUSE'),
                    ('station_setting', 'SS_STATION_SETTING'),
                    ('station_gaw_type', 'SG_STATION_GAW_TYPE'),
                    ('station_wmo_region', 'SW_STATION_WMO_REGION'),
                    ('station_latitude', 'ST_LATITUDE'),
                    ('station_longitude', 'ST_LONGITUDE'),
                    ('station_altitude', 'ST_ALTITUDE_ASL'),
                ]
                for keys in maps:
                    if keys[0] not in metadata:
                        metadata[keys[0]] = master_st[keys[1]]

    def _set_org(self, metadata):
        """
        Map the organisation code.
        """
        org = {
            # Empty org dict, just to make sure all keys exist.
            'OR_CODE': None,
            'OR_NAME': None,
            'OR_ACRONYM':  None,
            'OR_UNIT':  None,
            'OR_ADDR_LINE1':  None,
            'OR_ADDR_LINE2':  None,
            'OR_ADDR_ZIP':  None,
            'OR_ADDR_CITY':  None,
            'OR_ADDR_COUNTRY':  None,
        }
        if 'org' in metadata and metadata.org:
            if 'OR_CODE' in metadata.org and metadata.org.OR_CODE:
                try:
                    org.update(EbasMasterOR()[metadata.org['OR_CODE']])
                except KeyError:
                    self.warning(
                        "Organisation '{}' not found in EBAS "
                        "masterdata.".format(metadata.org['OR_CODE']))
        elif 'lab_code' in metadata and metadata.lab_code:
            self.warning(
                "No organisation defined in masterdata. Using lab code")
            try:
                org.update(EbasMasterOR()[metadata.lab_code])
            except KeyError:
                self.warning(
                    "Organisation '{}' not found in EBAS masterdata.".format(
                    metadata.lab_code))
        org.update(metadata.org)
        metadata.org = org


class EbasEanetFilterBase(EbasEanetBase):  # pylint: disable=W0223
    # W0223: Method ... is abstract in base class but is not overridden
    # This is also an abstract class, just does not raise NotImplementedError
    """
    Base class.
    """
    # set exception for generic read method
    READ_EXCEPTION = EbasEanetReadError

    def __init__(self, *args, **kwargs):
        """
        Class initioalization.
        """
        EbasEanetBase.__init__(self, *args, **kwargs)
        # add one additional attribute:
        self.sourcefile = None
        self.setup_non_domain()


class EbasEanetPrecipBase(EbasEanetBase):  # pylint: disable=W0223
    # W0223: Method ... is abstract in base class but is not overridden
    # This is also an abstract class, just does not raise NotImplementedError
    """
    Base class.
    """
    # set exception for generic read method
    READ_EXCEPTION = EbasEanetReadError

    def __init__(self, *args, **kwargs):
        """
        Class initioalization.
        """
        EbasEanetBase.__init__(self, *args, **kwargs)
        # add one additional attribute:
        self.sourcefile = None
        self.setup_non_domain()
        self._prec_values = None



"""
nilutility/logging_helper.py
$Id: logging_helper.py 2452 2020-04-23 19:20:47Z pe $

Module for logging initialisation.
This module is just a helper for the default python logging module.
It helps set up logging. But the use of python logging for the user modules
is completely uncahnged.
In the default case, the main script uses nilu.logging to set up the logging
facilities and other customizations. Logging itself will be uncahnges and
still done by using the default logging module

Features:
- typical logging setups can be done in a simple call to init_logging
- CustomFormatter class provides additional format names
  (level_on_>=warning, level+logger_on_>=warning)
- modified default log format for console output:
  '%(level+logger_on_>=warning)s%(message)s'
  (more practical for most usage cases)
- modified default log format for logfile output:
  '%(asctime)s %(name)-12s %(levelname)-8s: %(message)s'

Synopsis:

* general purpose console output (for usual commandline tools)
import logging
import nilu.logging_helper
nilu.logging_helper.init_logging(console={'level': logging.INFO})
logger = logging.getLogger('MyMainModule')
logger.info('hello')
logger.debug('debug test')
logger.warning('watch out!')
logger.critical('too late...')
>>CONSOLE OUTPUT:
hello
debug test
WARNING (MyMainModule): watch out!
CRITICAL (MyMainModule): too late...

* background task, silent, but create a logfile
import logging
import nilu.logging_helper
nilu.logging_helper.init_logging(\
    logfile=[{'level': logging.INFO, 'file': '/tmp/%Y%m%d.log'}])
logger = logging.getLogger('MyMainModule')
logger.info('hello')
logger.debug('debug test')
logger.warning('watch out!')
logger.critical('too late...')
>>CONSOLE OUTPUT:
>>/tmp/<date>.log:
2011-06-18 20:59:23,628 MyMainModule INFO    : hello
2011-06-18 20:59:23,631 MyMainModule WARNING : watch out!
2011-06-18 20:59:23,632 MyMainModule CRITICAL: too late...

* info and higher to console, one normal logfile and one debug log
import logging
import nilu.logging_helper
nilu.logging_helper.init_logging(\
    console={'level': logging.INFO},
    logfile=[{'level': logging.INFO, 'file': '/tmp/info_%Y%m%d.log'},
             {'level': logging.DEBUG,'file': '/tmp/debug_%Y%m%d.log'},])
logger = logging.getLogger('MyMainModule')
logger.info('hello')
logger.debug('debug test')
logger.warning('watch out!')
logger.critical('too late...')
>>CONSOLE OUTPUT:
hello
WARNING (MyMainModule): watch out!
CRITICAL (MyMainModule): too late...
>>/tmp/info_<date>.log:
2011-06-18 21:03:41,271 MyMainModule INFO    : hello
2011-06-18 21:03:41,275 MyMainModule WARNING : watch out!
2011-06-18 21:03:41,277 MyMainModule CRITICAL: too late...
>>/tmp/debug_<date>.log:
2011-06-18 21:03:41,271 MyMainModule INFO    : hello
2011-06-18 21:03:41,273 MyMainModule DEBUG   : debug test
2011-06-18 21:03:41,275 MyMainModule WARNING : watch out!
2011-06-18 21:03:41,277 MyMainModule CRITICAL: too late...

* play with format (console) and dateformats and timezones (logfiles):
import logging
import time
import nilu.logging_helper
nilu.logging_helper.init_logging(\
    console={'level': logging.INFO,
             'format': '%(name)-12s %(levelname)-8s: %(message)s'},
    logfile=[{'level': logging.DEBUG,
              'file': '/tmp/localtime_%Y%m%d%H%M.log',
              'datefmt': '%Y%m%d%H%M%S'},
             {'level': logging.DEBUG,
              'file': '/tmp/utc_%Y%m%d%H%M.log',
              'datefmt': '%Y%m%d%H%M%S',
              'datecvt': time.gmtime}])
logger = logging.getLogger('MyMainModule')
logger.info('hello')
logger.debug('debug test')
logger.warning('watch out!')
logger.critical('too late...')
>>CONSOLE OUTPUT:
MyMainModule INFO    : hello
MyMainModule WARNING : watch out!
MyMainModule CRITICAL: too late...
>>/tmp/localtime_<localtimestamp>.log:
20110618214534 MyMainModule INFO    : hello
20110618214534 MyMainModule DEBUG   : debug test
20110618214534 MyMainModule WARNING : watch out!
20110618214534 MyMainModule CRITICAL: too late...
>>/tmp/utc_<utctimestamp>.log:
20110618194534 MyMainModule INFO    : hello
20110618194534 MyMainModule DEBUG   : debug test
20110618194534 MyMainModule WARNING : watch out!
20110618194534 MyMainModule CRITICAL: too late...


History:
V.1.0.0  2011-06-17  pe  initial version

"""

import os
import sys
import logging
import logging.handlers
import time
import datetime
import getpass

INFO_PLUS = 25
INFO_MINUS = 15

class CustomFormatter(logging.Formatter):
    """
    Customized logging formatter class.

    Adiitional features:
        level_on_>=warning
            prints level name if level is >= WARNING
            e.g. "CRITICAL: <rest of log text>"
        level+logger_on_>=warning
            prints level and logger name if level is >= WARNING
            e.g. "CRITICAL (MyMainModule): <rest of log text>"
    """
    def format(self, record):
        """
        Format the specified record as text.

        fills:
            record.__dict__['level_on_>=warning']
            record.__dict__['level+logger_on_>=warning']
        and calls base class method to do the actual formatting.
        """
        if record.__dict__['levelno'] >= logging.WARN:
            record.__dict__['level_on_>=warning'] = logging.getLevelName(\
                record.levelno) + "): "
            record.__dict__['level+logger_on_>=warning'] = \
                logging.getLevelName(record.levelno) + \
                " (" + record.name + "): "
        else:
            record.__dict__['level_on_>=warning'] = ''
            record.__dict__['level+logger_on_>=warning'] = ''
        return logging.Formatter.format(self, record)

class CustomHandlerBase(object):
    """
    Provides functionality for custom handlers
    """
    def generate_filename(self, spec):
        """
        Generate the filename for logging handlers which log to files.
         - substitutes date/time format specifiers in filepath
         - option createdirs
        Parameters:
            spec   file handler specification (dict)
        Returns:
            file name
        """
        # the final datatype of now will be datetime.datetime in order
        # to to get microseconds support. (%f can be used in filename)

        # - now is set to spec[timestamp] if provided
        if 'timestamp' in spec:
            now = spec['timestamp']
            # timestamp may be struct_time or datetime
            if isinstance(now, time.struct_time):
                now = datetime.datetime.fromtimestamp(time.mktime(now))
        # - else now will be retrieved from datecvt if given
        elif 'datecvt' in spec:
            now = spec['datecvt']()
            # datecvt() must return struct_time object (generally for logging)
            # so this is expected to always be true:
            if isinstance(now, time.struct_time):
                now = datetime.datetime.fromtimestamp(time.mktime(now))
        # - else set now to datetime object right away
        else:
            now = datetime.datetime.now()

        # substitute dat/time format specifiers
        filename = now.strftime(spec['file'])
        # expand ~ to users directories ~ home, ~user other user's home:
        filename = os.path.expanduser(filename)
        # substitute $user$
        if '$user$' in filename:
            filename = filename.replace('$user$', getpass.getuser())
        if 'createdirs' in spec and spec['createdirs']:
            filepath = os.path.dirname(filename)
            if filepath and not os.path.isdir(filepath):
                os.makedirs(filepath, 0o775)
        return filename

class CustomFileHandler(logging.FileHandler, CustomHandlerBase):
    """
    File handler with some extra functionality:
     - substitutes date/time format specifiers in filepath
     - option createdirs

    The handler can pass an extra keys from spec to the underlying
    logging.FileHandler:
        mode (deafult 'a')
    """
    def __init__(self, spec):
        """
        Setup customized FileHandler.
        Parameters:
            spec   file handler specification (dict)
        Returns:
            None
        """
        filename = self.generate_filename(spec)
        mode = 'a'  # default: append to file when exists
        if 'mode' in spec:
            mode = spec['mode']
        super(CustomFileHandler, self).__init__(filename, mode)
        self.setLevel(spec['level'])

class CustomTimedRotatingFileHandler(logging.handlers.TimedRotatingFileHandler,
                                     CustomHandlerBase):
    """
    TimedRotatingFileHandler with some extra functionality:
     - substitutes date/time format specifiers in filepath
     - option createdirs

    The handler can pass some extra keys from spec to the underlying
    logging.TimedRotatingFileHandler:
        when (defult 'h'), interval (default 1), backupCount (default 0),
        utc (default False)
    """
    def __init__(self, spec):
        """
        Setup customized TimedRotatingFileHandler
        Parameters:
            spec   handler specification (dict)
        Returns:
            None
        """
        filename = self.generate_filename(spec)
        kwlist = ('when', 'interval', 'backupCount', 'encoding', 'delay', 'utc')
        kwargs = {kw: spec[kw] for kw in kwlist if kw in spec}
        super(CustomTimedRotatingFileHandler, self).__init__(filename, **kwargs)
        self.setLevel(spec['level'])

class StreamLogger(object):
    # pylint: disable=R0903
    #  R0903: Too few public methods
    """
    Stream object which redirects output to a logger.
    (use this to redirect streams like stdout, stderr)
    """
    def __init__(self, logger, loglevel=logging.INFO):
        """
        Set up StreamLogger object.
        Parameters:
            logger    logger object
            loglevel  log level to be used for messages through this stream
        """
        self.logger = logger
        self.loglevel = loglevel

    def write(self, buf):
        """
        Write method for stream. Writes to logger.
        Parameters:
            buf   output string
        Returns:
            None
        """
        for line in buf.rstrip().splitlines():
            self.logger.log(self.loglevel, line)


def init_logging(console=None, logfile=None, loggers=None, catch_stdout=False,
                 catch_stderr=False):
    """
    Initialize logging.
    Sets up all needed logging facilities and logger configurations.

    Parameters:
       console     definition for console logging
                   dictionary:
                      level:  minimum loglevel for console output
                      format: format for logging
                      datefmt: date format (format time.strftime())
                      datecvt: date converter function
                               Control the timezone for output.
                               e.g. time.gmtime, time.localtime [=default]

                   e.g.:
                      {'level': logging.DEBUG,
                       'format': '%(level+logger_on_>=warning)s%(message)s'}
       logfile     array of dictionaries (multiple logfiles possible):
                      file:    file name (may contain time format elements
                               like %Y%m%d)
                      level:   minimum loglevel for console output
                      format:  format for logging
                      datefmt: date format (format time.strftime())
                      datecvt: date converter function
                               This affects format strings for the log file
                               name as well as the log entries timestamp
                               e.g. time.gmtime, time.localtime [=default]
                      timestamp:
                               provide a timestamp for filename generation
                               (time or datetime.datetime)
                   e.g.:
                      [{'file': 'log', 'level' logging.DEBUG }, ...]
       loggers     minimum loglevels per logger name (modules?)
                   If not defined, all loggers will inherit their level from
                   the root logger.
                   array of dictionaries:
                      [{'<logger-name>': <level>}, ...}
       catch_stdout
                   if true, stdout stream will be caught and written as
                   to logger name STDOUT and severity INFO
       catch_stderr
                   if true, stderr stream will be caught and written as
                   to logger name STDERR and severity ERROR

    Returns:
       None
    """
    logging.addLevelName(25, 'INFO+')
    logging.addLevelName(15, 'INFO-')

    _init_logging_console(console)
    _init_logging_logfiles(logfile)

    if loggers is None:
        # should not use {} as default argument value (pylint W:158)
        loggers = {'root': logging.DEBUG}
    logging.root.level = loggers['root']
    for log in list(loggers.keys()):
        if log != 'root':
            logger = logging.getLogger(log)
            logger.level = loggers[log]

    if catch_stdout:
        stdout_logger = logging.getLogger('STDOUT')
        slo = StreamLogger(stdout_logger, logging.INFO)
        sys.stdout = slo
    if catch_stderr:
        stderr_logger = logging.getLogger('STDERR')
        slo = StreamLogger(stderr_logger, logging.ERROR)
        sys.stderr = slo

def _init_logging_console(console=None):
    """
    Initialize logging. Set up log handler for console output.

    Parameters:
       console     definition for console logging
                   dictionary:
                      level:  minimum loglevel for console output
                      format: format for logging
                      datefmt: date format (format time.strftime())
                      datecvt: date converter function
                               Control the timezone for output.
                               e.g. time.gmtime, time.localtime [=default]

                   e.g.:
                      {'level': logging.DEBUG,
                       'format': '%(level+logger_on_>=warning)s%(message)s'}

    Returns:
       None
    """
    if console != None:
        if not 'level' in console:
            console['level'] = logging.INFO
        if console['level'] <= logging.CRITICAL:
            consolehdl = logging.StreamHandler()
            consolehdl.setLevel(console['level'])
            if 'format' in console and console['format'] != None:
                formatter_c = CustomFormatter(console['format'])
            else:
                formatter_c = CustomFormatter(\
                    '%(level+logger_on_>=warning)s%(message)s')
            if 'datefmt' in console:
                formatter_c.datefmt = console['datefmt']
            if 'datecvt' in console:
                formatter_c.converter = console['datecvt']
            consolehdl.setFormatter(formatter_c)
            logging.root.addHandler(consolehdl)

def _init_logging_logfiles(logfiles=None):
    """
    Initialize logging. Set up log handler for logfile(s)

    Parameters:
       logfiles    array of dictionaries (multiple logfiles possible):
                      handler  CustomFileHandler, CustomTimedRotatingFileHandler
                      file:    file name (may contain time format elements
                               like %Y%m%d)
                      level:   minimum loglevel for console output
                      format:  format for logging
                      datefmt: date format (format time.strftime())
                      datecvt: date converter function
                               This affects format strings for the log file
                               name as well as the log entries timestamp
                               e.g. time.gmtime, time.localtime [=default]
                      timestamp:
                               provide a timestamp for filename generation
                               (time or datetime.datetime)
                      some handlers have extra attributes:
                          CustomFileHandler:
                              mode: file mode, default 's'
                          CustomTimedRotatingFileHandler:
                              when: rotating time S, M, D, W0-W6, midnight
                              interval: how many 'when's to rotate
                              backupCount: how many files to kkep
                              utc: genera filename for backup files in utc
                   e.g.:
                      [{'file': 'log', 'level' logging.DEBUG }, ...]
    """
    filehdls = []
    if logfiles is None:
        # should not use [] as default argument value (pylint W:158)
        logfiles = []
    for log in logfiles:
        if not 'level' in log:
            log['level'] = logging.INFO
        if log['level'] <= logging.CRITICAL:
            if not 'file' in log or log['file'] is None:
                raise ValueError('insufficient logfile definition')
            if 'createdirs' not in log:
                log['createdirs'] = True
            if 'handler' in log:
                logfilehdl = log['handler'](log)
            else:
                logfilehdl = CustomFileHandler(log)
            formatter_f = None
            if 'format' in log:
                formatter_f = CustomFormatter(log['format'])
            else:
                formatter_f = CustomFormatter(\
                    '%(asctime)s %(name)-12s %(levelname)-8s: %(message)s')
            if 'datefmt' in log:
                formatter_f.datefmt = log['datefmt']
            if 'datecvt' in log:
                formatter_f.converter = log['datecvt']
            logfilehdl.setFormatter(formatter_f)
            filehdls.append(logfilehdl)
    for filehdl in filehdls:
        logging.root.addHandler(filehdl)



"""
EBAS Masterdata Class for absorption_crosssection metadata

This module implements the class EbasMasterAX.
"""

from .offline_masterdata import AXOfflineMasterData
from .base import EbasMasterBase

class EbasMasterAX(EbasMasterBase, AXOfflineMasterData):
    """
    Domain Class for analyticalinstrument type masterdata
    Objects of this class do not represent entities, this class provides class
    methods for handling components and checking them against master data.
    Components master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for AX (ax_absorption_crosssection)
    # Those are fallback values, will be read from database as soon as possible.
    AXOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read at masterdata if dbh is provided.
        """
        AXOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)

    @classmethod
    def exist_absorption_crosssection(cls, absorption_crosssection):
        """
        Check if the absorption crosssection exists (regardless of CO).
        Parameters:
            absorption_crosssection    absorption crosssection code to be checked
        Returns:
            True/False
        """
        if any([key[1] == absorption_crosssection
                for key in list(cls.META.keys())]):
            return True
        return False



"""
$Id: postgres.py 2599 2021-01-27 16:27:04Z pe $

Helper module fro accessing PostgreSQL databases.

# Synopsis:

## Build an application db access class

Advantage: convert strings to unicode, log executes as debug, make use of
           _convert_row2dict and _convert_rows2dictlist

```python
class MyDbAccess(PostgresDbBase):
    def seltestdata(self):
        curs = self.conn.cursor()
        curs.execute('select * from test')
        return self._convert_rows2dictlist(curs.fetchall(), curs.description)

...

db = MyDbAccess(host=host, user=user, password=password, dbname=dbname)
print db.seltestdata()
print db.seltestdata()[1].columnname
```

## Use customized classes w/o creating a dedicated db layer class:

e.g. for small scripts.
Advantage: convert strings to unicode, log executes as debug.

```python
import databases.postgres
import psycopg2

con = psycopg2.connect(
    host=None, user='paul', password=None, dbname='paul',
    connection_factory=databases.postgres.PostgresDbConnection,
    cursor_factory=databases.postgres.PostgresDbCursor)
curs = con.cursor()
curs.execute('select * from test')
print list(curs.fetchall())
```
"""

import logging
import uuid
from textwrap import dedent
import psycopg2
from psycopg2 import Error as DbError  # for import in clients
from psycopg2.extras import execute_values
import six
import re
import datetime
import psycopg2.extensions
from nilutility.datetime_helper import DatetimeInterval
from nilutility.datatypes import DataObject
from .base import DbBase

# make sure strings are converted form database character set to unicode
# in python2 (in python3 it's default)
psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY)

class PostgresDbConnection(psycopg2.extensions.connection):
    # pylint: disable=R0903
    # R0903: Too few public methods
    """
    Derived class for a postgeres connection object.
    """

    def commit(self):
        """
        Log commit with severity debug.
        """
        logger = logging.getLogger('PostgresDb')
        logger.debug(
            'commit transaction (%s)', self.appname)
        psycopg2.extensions.connection.commit(self)

    def rollback(self):
        """
        Log commit with severity debug.
        """
        logger = logging.getLogger('PostgresDb')
        logger.debug(
            'rollback transaction (%s)', self.appname)
        psycopg2.extensions.connection.rollback(self)

    @property
    def appname(self):
        """
        Get the application_name from dsn_parameters, None as fallback
        """
        try:
            return self.get_dsn_parameters()['application_name']
        except KeyError:
            return None


class PostgresDbCursor(psycopg2.extensions.cursor):  # pylint: disable=R0903
    # R0903: Too few public methods
    """
    Logging cursor object. Logs every execute with severity debug, logs
    exceptions as error.
    """
    def execute(self, query, vars_=None):
        """
        Log execute with severity debug, log exceptions as error.
        Parameters
        """
        logger = logging.getLogger('PostgresDb')
        logger.debug(u'execute: %s', self.mogrify(query, vars_))
        logger.debug(u'execute: %s',
            self.mogrify(query, vars_).decode(self.connection.encoding))
        try:
            psycopg2.extensions.cursor.execute(self, query, vars_)
        except Exception as exc:
            logger.error("%s: %s", exc.__class__.__name__, exc)
            raise

    def executemany(self, query, vars_list):
        """
        Log executemany with severity debug, log exceptions as error.
        Parameters
        """
        logger = logging.getLogger('PostgresDb')
        logger.warning(
            'executemany method is not performing well, consider using '
            'the execute_values method')
        if logger.getEffectiveLevel() <= logging.DEBUG:
            logger.debug('executemany: %s', query)
            for _, vars_ in enumerate(vars_list):
                if isinstance(vars_, dict):
                    logger.debug(
                        'executemany vars: %s',
                        ', '.join(
                            ['{0}: {1} ({2})'.format(key, value, type(value))
                             for (key, value) in list(vars_.items())]))
                else:
                    logger.debug('executemany vars_list: %s', vars_)
        try:
            psycopg2.extensions.cursor.executemany(self, query, vars_list)
        except Exception as exc:
            logger.error("%s: %s", exc.__class__.__name__, exc)
            raise

    def execute_values(self, sql, argslist, template=None, page_size=100,
                       fetch=False):
        """
        Log execute_values with severity debug, log exceptions as error.
        Parameters:
            sql        the query to execute.
                       It must contain a single %s placeholder, which will be
                       replaced by a VALUES list.
            argslist   sequence of sequences or dictionaries with the arguments
                       to send to the query.
            template   the snippet to merge to every item in argslist to compose
                       the query.
                       (e.g. "(%s, %s, %s)", or "(%s, %s, 42)" for const values)
                       If not specified, assume the arguments are sequence and
                       use a simple positional template (i.e. (%s, %s, ...))
            page_size  maximum number of argslist items to include in every
                       statement. If there are more items the function will
                       execute more than one statement.
            fetch      if True return the query results into a list (like in a
                       fetchall()). Useful for queries with RETURNING clause.
        Returns:
            None or list
        """
        logger = logging.getLogger('PostgresDb')
        if logger.getEffectiveLevel() <= logging.DEBUG:
            logger.debug('execute_values: %s', sql)
            logger.debug('execute_values template: %s', template)
            logger.debug('execute_values page_size: %s', page_size)
            logger.debug('execute_values fetch: %s', fetch)
            for _, args in enumerate(argslist):
                if isinstance(args, dict):
                    logger.debug(
                        'execute_values args: %s',
                        ', '.join(
                            ['{0}: {1} ({2})'.format(key, value, type(value))
                             for (key, value) in list(args.items())]))
                else:
                    logger.debug('execute_values args: %s', args)
        try:
            execute_values(self, sql, argslist, template=template,
                           page_size=page_size, fetch=fetch)
        except Exception as exc:
            logger.error("%s: %s", exc.__class__.__name__, exc)
            raise


class PostgresDbBase(DbBase):
    """
    Base class for application database layer classes (should be derived from
    this class).
    """

    # Used by base class
    DB_MAXLEN_IDENTIFIER = 63

    def __init__(self):
        """
        Initialize DB connection. Open DB.
        Parameters:
            all parameters accepted by psycopg2.connect
            most prominent:
                dbname, user, password, host, port
        Returns:
            None
        Additional parameters can be either suplied by keyword parameters or a
        dsn connection sting. See:
        https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-CONNSTRING
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cursor_fetch_buffer = 1000
        self.tmp_addsql_crit_tables = []
        self.conn = None

    def _connect(self, connection_factory=None, cursor_factory=None, **kwargs):
        """
        Connect the database.
        Parameters:
            all parameters accepted by psycopg2.connect
        Returns:
            connection   can't sett self.conn (must also work for additional
                         conn's (e.g. EbasDb.conn_ts)
                         thus, the conn is returned and self.xxx is set by the
                         caller
        Additional parameters can be either suplied by keyword parameters or a
        dsn connection sting. See:
        https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-CONNSTRING
        """
        if connection_factory is None:
            connection_factory = PostgresDbConnection
        if cursor_factory is None:
            cursor_factory = PostgresDbCursor
        conn = psycopg2.connect(
            connection_factory=connection_factory,
            cursor_factory=cursor_factory, **kwargs)
        self.logger.debug('connected database %s', conn.get_dsn_parameters())
        # Set the default timezone to UTC
        # The database default is also set to UTC, but to be sure, in case the
        # user set something different on the client side (e.g. env var PGTZ).
        curs = conn.cursor()
        curs.execute("set timezone to 'Etc/UTC'")
        curs.execute("set datestyle to 'ISO'");
        self._typecast_adapt_utc(conn)
        return conn

    def close(self):
        """
        Close the db connection, cleanup.
        """
        self.conn.close()

    def commit(self):
        """
        Commit changes in the connection.
        """
        # Make sure the temporary entries in TMP_ADDSQL_CRIT do not persist
        self._cleanup_addsql_crit()
        self.conn.commit()


    def check_constraints(self):
        """
        Check deferred constraints and constraint triggers.
        (Can be checked in case of no commit)
        """
        self._execute('SET CONSTRAINTS ALL IMMEDIATE')

    def rollback(self):
        """
        Rollback changes in the connection.
        """
        self.conn.rollback()

    def _execute(self, query, vars_=None):
        """
        Provides an execute method for the object (private, only to be used
        internally by derived object's access methods).
        Open a new cursor, execute and return the cursor.
        Includes logging through the customized PostgresDbCursor object.
        Parmeter:
            query    the sql query
            vars_    variables for the query
        Returns:
            cursor object
        """
        curs = self.conn.cursor()
        curs.execute(query, vars_)
        return curs

    def _executemany(self, query, vars_list):
        """
        Provides an executemany method for the object (private, only to be used
        internally by derived object's access methods).
        Open a new cursor and executemany on it.
        Includes logging through the customized PostgresDbCursor.
        WARNING: executemany does not perform well. Use _execute_values for
        better performance if possible (needs different parameters).

        Parmeter:
            query      the sql query
            vars_list  list of variables for the query
        Returns:
            None
        """
        if self.conn.cursor_factory != PostgresDbCursor:
            self.logger.warning(
                'executemany method is not performing well, consider using '
                'PostgresDbCursor as cursor_factoy and the execute_values '
                'method')
        curs = self.conn.cursor()
        curs.executemany(query, vars_list)

    def _execute_values(self, sql, argslist, template=None, page_size=100,
                        fetch=False):
        """
        Provides an execute_values method for the object (private, only to be
        used internally by derived object's access methods).
        Open a new cursor and execute_values on it.
        Includes logging through the customized PostgresDbCursor and is faster
        than the executemany method.

        Parameters:
            sql        the query to execute.
                       It must contain a single %s placeholder, which will be
                       replaced by a VALUES list.
            argslist   sequence of sequences or dictionaries with the arguments
                       to send to the query.
            template   the snippet to merge to every item in argslist to compose
                       the query.
                       (e.g. "(%s, %s, %s)", or "(%s, %s, 42)" for const values)
                       If not specified, assume the arguments are sequence and
                       use a simple positional template (i.e. (%s, %s, ...))
            page_size  maximum number of argslist items to include in every
                       statement. If there are more items the function will
                       execute more than one statement.
            fetch      if True return the query results into a list (like in a
                       fetchall()). Useful for queries with RETURNING clause.
        Returns:
            None or list
        """
        curs = self.conn.cursor()
        curs.execute_values(sql, argslist, template=template,
                            page_size=page_size, fetch=fetch)
        # Exception thrown if curs has no execute_values method, i.e. not a
        # PostgresDbCursor

    def _yield_dict(self, curs):
        """
        Yields the cursur results as dictionary per row.

        Parameters:
            curs    opened cursor object
        Returns:
            generator, dictionaries with results, None if no result
        """
        while True:
            results = curs.fetchmany(self.cursor_fetch_buffer)
            if not results:
                curs.close()
                break
            for result in results:
                yield self._convert_row2dict(result, curs.description)

    @staticmethod
    def _convert_row2dict(result, desc):
        """
        Converts a fetched row (tupel) to a dictionary (keys=colnames).

        Parameters:
            result    fetched result row (tupel)
            desc      cursor description
        """
        ret = DataObject()
        for i, item in enumerate(result):
            colnam = desc[i][0]
            ret[colnam] = item
        return ret

    @classmethod
    def _convert_rows2dictlist(cls, resultlist, desc):
        """
        Converts a fetched row (tupel) to a dictionary (keys=colnames).

        Parameters:
            resultlist  fetched result rows (list of tupel)
            desc        cursor description
        """
        ret = []
        for row in resultlist:
            ret.append(cls._convert_row2dict(row, desc))
        return ret

    @staticmethod
    def _hostvar_sql(varname):
        """
        Returns the sql part of a host variable name.
        """
        return "%({})s".format(varname)

    @staticmethod
    def _hostvar_param(varname):
        """
        Returns the parameter dictionary key for a host variable.
        """
        return varname

    def _insert_addsql_crit(self, in_list, alt_curs=None):
        """
        Insert table entries for addsql_crit.
        Parameters:
            in_list    list of criteria values
            alt_cur    alternative cursor to be used (special case for EbasDb)
        Returns:
            temp_id    id for the values in the temp table
        """
        if any([isinstance(elem[1], six.string_types) for elem in in_list]):
            valtype = 'VAL_CHR'
        elif any([isinstance(elem[1], float) for elem in in_list]):
            valtype = 'VAL_DBL'
        else:
            valtype = 'VAL_INT'
        temp_sql = dedent("""\
             insert into tmp_addsql_crit (ID, TIME, {})
                 values (%(ID)s, current_timestamp AT TIME ZONE 'UTC',
                 %(VAL)s)""".format(valtype))
        temp_id = uuid.uuid1().hex
        self.tmp_addsql_crit_tables.append(temp_id)
        temp_params = {'ID': temp_id}
        for elem in in_list:
            temp_params['VAL'] = elem[1]
            if alt_curs:
                alt_curs.execute(temp_sql, temp_params)
            else:
                self._execute(temp_sql, temp_params)
        return "select {} from tmp_addsql_crit where ID='{}'".format(
            valtype, temp_id)

    def _cleanup_addsql_crit(self, alt_curs=None):
        """
        Cleanup temp table entries from addsql_crit
        Parameters:
            alt_cur    alternative cursor to be used (special case for EbasDb)
        Returns:
            None
        """
        sql = "delete from tmp_addsql_crit where ID=%(ID)s"
        params = {}
        for id_ in self.tmp_addsql_crit_tables:
            params['ID'] = id_
            if alt_curs:
                alt_curs.execute(sql, params)
            else:
                self._execute(sql, params)
        self.tmp_addsql_crit_tables = []


    ###
    ### timestamp related adapters and casters
    ###
    ### By default, all timestamps will be naive datetime.datetime objects
    ### (timezone unaware). All times are UTC. Always! So no use in carrying
    ### the load of timezone through the code.
    ###
    ### Maybe some application will need alternative implementations...
    ###
    ### The postgres db timestamps are mostly timestamptz, i.e. including
    ### timezone. The db module adapts and casts the types accordingly.

    # compile the regexp just once for performance:
    REG_TSTZ = re.compile(
        r'^(\d\d\d\d)-(\d\d)-(\d\d) (\d\d):(\d\d):(\d\d)(\.\d*)?\+00$')
    REG_TSTZRANGE = re.compile(r"""
        ( \(|\[ )                   # lower bound flag
        (?:                         # lower bound:
          " ( (?: [^"] | "")* ) "   #   - a quoted string
          | ( [^",]+ )              #   - or an unquoted string
        )?                          #   - or empty (not catched)
        ,
        (?:                         # upper bound:
          " ( (?: [^"] | "")* ) "   #   - a quoted string
          | ( [^"\)\]]+ )           #   - or an unquoted string
        )?                          #   - or empty (not catched)
        ( \)|\] )                   # upper bound flag
        """, re.VERBOSE)

    def _typecast_adapt_utc(self, conn):
        """
        Set up all adapters and type casters.
        Parameters:
            conn    connection object (can't use self.conn here, as this must
                    possibly work with other connections too (e.g.
                    EbasDb.conn_ts)
        """

        def cast_timestamptz(value, curs):
            """
            Cast a postgres timstamptz to a naive datetime.datetime.
            """
            if value is None:
                return None
            reg = re.match(self.__class__.REG_TSTZ, value)
            frag = 0 if not reg.group(7) \
                else int(round(float(reg.group(7))*1000000.0))
            return datetime.datetime(
                int(reg.group(1)), int(reg.group(2)), int(reg.group(3)),
                int(reg.group(4)), int(reg.group(5)), int(reg.group(6)),
                frag)

        def adapt_dt(dt_):
            """
            Adapter for datetime.datetime (to timestamptz).
            """
            return psycopg2.extensions.AsIs(
                dt_.strftime("'%Y-%m-%d %H:%M:%S.%f+00'::timestamptz"))

        def cast_tstzrange(value, curs):
            """
            Cast a postgres tstzrange to a DatetimeInterval with naive datetime
            elements.
            """
            if value == 'empty':
                return DatetimeInterval.empty()
            reg = re.match(self.__class__.REG_TSTZRANGE, value)
            return DatetimeInterval(
                cast_timestamptz(reg.group(2), curs) if reg.group(2) else
                cast_timestamptz(reg.group(3), curs) if reg.group(3) else None,
                cast_timestamptz(reg.group(4), curs) if reg.group(4) else
                cast_timestamptz(reg.group(5), curs) if reg.group(5) else None,
                bounds=reg.group(1) + reg.group(6))

        def adapt_dtint(dt_):
            """
            Adapter for DatetimeInterval (to tstzrange).
            """
            if dt_.is_empty():
                return psycopg2.extensions.AsIs("'empty'::tstzrange")
            return psycopg2.extensions.AsIs(
                """'{}{},{}{}'::tstzrange""".format(
                    dt_.bounds[0],
                    '"{}"'.format(dt_[0].strftime('%Y-%m-%d %H:%M:%S.%f+00')) \
                        if dt_[0] else '',
                    '"{}"'.format(dt_[1].strftime('%Y-%m-%d %H:%M:%S.%f+00')) \
                        if dt_[1] else '',
                    dt_.bounds[1]))

        curs = conn.cursor()
        curs.execute("select current_timestamp")
        tstztype = psycopg2.extensions.new_type(
            (curs.description[0][1],), "timestamptz", cast_timestamptz)
        # Register type conversion, limit scope to only the two ebas.db
        # connections. Otherwise problems with other code parts which use
        # another database interface.
        psycopg2.extensions.register_type(tstztype, conn)

        psycopg2.extensions.register_adapter(datetime.datetime, adapt_dt)

        curs = conn.cursor()
        curs.execute("select '[2021-01-01, 2022-01-01)'::tstzrange")
        tstzrangetype = psycopg2.extensions.new_type(
            (curs.description[0][1],), "tstzrange", cast_tstzrange)
        # Register type conversion, limit scope to only the two ebas.db
        # connections. Otherwise problems with other code parts which use
        # another database interface.
        psycopg2.extensions.register_type(tstzrangetype, conn)

        psycopg2.extensions.register_adapter(DatetimeInterval, adapt_dtint)


class PostgresDb(PostgresDbBase):
    """
    Final class for database access in Postgres. Using mainly the execute method
    (client needs to formulate the sql).
    This class is meant to be used for more ad-hoq dataabase queries. Derived
    classes which aim to provide a set of customized db access methods for an
    application, should be derived from PostgresDbBase instead of this class.
    See e.g. ebas/db4 for amore advanced final implementation.
    """

    def __init__(self, **kwargs):
        """
        Initialize DB connection. Open DB.
        Parameters:
            all parameters accepted by psycopg2.connect
            most prominent:
                dbname, user, password, host, port
        Returns:
            None
        Additional parameters can be either suplied by keyword parameters or a
        dsn connection sting. See:
        https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-CONNSTRING
        """
        PostgresDbBase.__init__(self)
        self.conn = self._connect(**kwargs)

    def execute(self, *args, **kwargs):
        """
        Wrapper method to expose base classes _execute as public.
        See PostgresDbBase._execute

        Open a new cursor, execute and return the cursor.
        Includes logging through the customized PostgresDbCursor object.
        Parmeter:
            query    the sql query
            vars     variables for the query
        Returns:
            cursor object
        """
        return self._execute(*args, **kwargs)

    def executemany(self, *args, **kwargs):
        """
        Wrapper method to expose base classes _executemany as public.
        See PostgresDbBase._executemanty
        Open a new cursor and executemany on it.
        Includes logging through the customized PostgresDbCursor object.
        Parmeter:
            query      the sql query
            vars_list  list of variables for the query
        Returns:
            None
        """
        self._executemany(*args, **kwargs)

    def execute_values(self, *args, **kwargs):
        """
        Wrapper method to expose base classes _execute_values as public.
        See PostgresDbBase._execute_values
        Open a new cursor and execute_values on it.
        Includes logging through the customized PostgresDbCursor object.
        Parameters:
            sql        the query to execute.
                       It must contain a single %s placeholder, which will be
                       replaced by a VALUES list.
            argslist   sequence of sequences or dictionaries with the arguments
                       to send to the query.
            template   the snippet to merge to every item in argslist to compose
                       the query.
                       (e.g. "(%s, %s, %s)", or "(%s, %s, 42)" for const values)
                       If not specified, assume the arguments are sequence and
                       use a simple positional template (i.e. (%s, %s, ...))
            page_size  maximum number of argslist items to include in every
                       statement. If there are more items the function will
                       execute more than one statement.
            fetch      if True return the query results into a list (like in a
                       fetchall()). Useful for queries with RETURNING clause.
        Returns:
            None or list
        """
        self._execute_values(*args, **kwargs)



"""
/NasaAmes1001/write.py
$Id: write.py 2566 2020-12-09 13:11:46Z pe $

Basic NASA Ames 1001 file writer class.
Handles low level file I/O.

History:
V.1.0.0  2013-10-24  pe  initial version

"""

import datetime
import re
import sys
import io
from six import PY2, string_types
from math import log
from nilutility.ndecimal import Decimal
from nilutility.datatypes import HexInt
from nilutility.numeric import digits_stat, digits
from .base import NasaAmes1001Error, NasaAmes1001Base

class NasaAmes1001Write(NasaAmes1001Base):
    """
    Partial class NasaAmes1001: Part for writing NasaAmes1001 files from object.
    """
    def write(self, filespec=None, encoding='utf-8'):
        """
        Reads the file according to NasaAmes 1001 standard.
        Parameters:
            filespec    file name (incl. path) or file like object (stream)
                        stdout if not passed
            encoding    unicode encoding for output
        Returns:
            None
        Raises:
            IOError           file open error
            NasaAmesError     on any unrecoverable write error
        """
        # reset the object state for writing
        self.init_state()
        self.finalize_data()

        self.encoding = encoding
        if filespec is None:
            if PY2:
                self.file = sys.stdout
            else:  # PY3
                from io  import _io
                if isinstance(sys.stdout, _io.TextIOWrapper):
                    # std stream in py3 is an encoding wrapper. We need to use
                    # the unencoded buffer directly in order to write bytes.
                    # We need to check if it's really a TextIOWrapper, otherwise
                    # we get problems with jupyter.
                    self.file = sys.stdout.buffer
                else:
                    self.file = sys.stdout
        elif isinstance(filespec, string_types):
            filename = filespec
            self.logger.info("writing file {}".format(filename))
            if PY2:
                mod = "w"
            else:
                mod = "wb"
            with open(filename, mod) as self.file:
                self._write()
        else:
            filename = None
            self.logger.debug("writing to stream")
            self.file = filespec
            self._write()

    def finalize_data(self):
        """
        Calculate data attributes that are not setable by the caller (not in
        SETABLE_DATA_ATTRIBS). These would be redundant if set by the caller.
        Set default values for unset attributes (initialized to None).
        Parameters:
            None
        Returns:
            None
        Raises:
            various builtin exceptions when data are not set correctly.
        """
        # first set DATA, SCOM and NCOM defaults, if not set
        if self.data.DATA is None:
            self.data.DATA = []
        if self.data.SCOML is None:
            self.data.SCOML = []
        if self.data.NCOML is None:
            self.data.NCOML = []
        # calculate unsetable attributes
        self.data.NV = max(len(self.data.DATA) - 1, 0)
        self.data.NLHEAD = \
            14 + self.data.NV + len(self.data.SCOML) + len(self.data.NCOML)

        self.variable_stat() # sets VMISS, VFORMAT, VREPR
        self.data.DX = self.dx_from_data()

        # set default values for the rest of unset attributes
        if self.data.IVOL is None:
            self.data.IVOL = 1
        if self.data.NVOL is None:
            self.data.NVOL = 1
        if self.data.DATE is None:
            raise NasaAmes1001Error('DATE must be set for writing NasaAmes1001')
        if self.data.RDATE is None:
            self.data.RDATE = datetime.datetime.utcnow()
        if self.data.VSCAL is None:
            self.data.VSCAL = [1 for _ in range(self.data.NV)]
        if self.data.VNAME is None:
            self.data.VNAME = ['' for _ in range(self.data.NV)]
        self._checkdims()

    def _checkdims(self):
        """
        Checks consistency of dimensions and extents in the data structure.
        Parameters:
            None
        Returns:
            None
        Raises:
            NasAmes1001Error if inconsistent
        """
        # check dimension for VSCAL, VMISS, VNAME and DATA
        if len(self.data.VSCAL) != self.data.NV:
            raise NasaAmes1001Error(
                'VSCAL ({} values) inconsistent with NV ({})'
                .format(len(self.data.VSCAL), self.data.NV))
        if len(self.data.VMISS) != self.data.NV:
            raise NasaAmes1001Error(
                'VMISS ({} values) inconsistent with NV ({})'
                .format(len(self.data.VMISS), self.data.NV))
        if len(self.data.VNAME) != self.data.NV:
            raise NasaAmes1001Error(
                'VNAME ({} values) inconsistent with NV ({})'
                .format(len(self.data.VNAME), self.data.NV))
        if len(self.data.DATA)-1 != self.data.NV and \
           (self.data.NV != 0 or len(self.data.DATA) != 0):
            # rais error if #vars - 1 (== # dependent vars) != NV
            # special case: NV=0 and #vars = 0 (=> # dependent vars is -1!)
            raise NasaAmes1001Error(
                'DATA ({} dep. variables) inconsistent with NV ({})'
                .format(len(self.data.DATA)-1, self.data.NV))
        # check variable length
        for i in range(self.data.NV):
            if len(self.data.DATA[0]) != len(self.data.DATA[i+1]):
                raise NasaAmes1001Error(
                    'number of rows for variable[0] ({}) is different from '
                    'variable[{}] ({})'.format(len(self.data.DATA[0]), i+1,
                                               len(self.data.DATA[i+1])))

    def variable_stat(self):
        """
        Creates variable statistics.

        Parameters:
            var_index   variable index for which the statistics should be
                        gathered. None=all variables
        Returns:
            None

        These parameters are gathered/set:
         - VMISS: missing value
         - VFORMAT: output format for values
         - VREPR: data type for data variable
         - xformat
         - xmindig
         - minxmindig  (used when set from caller, not set)
         - maxxmindig  (used when set from caller, not set)
         - minvmindig  (used when set from caller, not set)
         - maxvmindig  (used when set from caller, not set)
        HexInt variables are checked for type
        """
        (self.data.VMISS, self.data.VFORMAT, self.data.VREPR) = ([], [], [])
        for i in range(len(self.data.DATA)):
            validvals = [x for x in self.data.DATA[i] if x is not None]
            if validvals:
                vrepr = type(validvals[0])
            else:
                vrepr = None
            (minval, maxval, mindig, maxdig, maxprec) = digits_stat(validvals)
            if mindig is None:
                # mindig and maxdig is None if all values are None
                # We handle this as if all values were 0
                (mindig, maxdig) = (0, 0)
#             if maxdig == 0:
#                 maxdig = 1
#             if mindig == 0:
#                 mindig = 1
#             if maxprec == 0:
#                 maxprec = 1
            if i == 0:
                # independent variable (start_time)
                if self.minxmindig != None and mindig < self.minxmindig:
                    mindig = self.minxmindig
                if self.maxxmindig != None and mindig > self.maxxmindig:
                    mindig = self.maxxmindig
                self.xmindig = mindig
                self.xformat = self.gen_vformat(vrepr, mindig, maxdig, maxprec,
                                                minval, True)
            else:
                # dependent variables (first is usually end_time)
                # collect some config parameters for this variable...
                vpar_minvmiss = False
                if self.minvmiss and len(self.minvmiss) >= i:
                    vpar_minvmiss = self.minvmiss[i-1]
                vpar_minvmindig = None
                if self.minvmindig and len(self.minvmindig) >= i:
                    vpar_minvmindig = self.minvmindig[i-1]                
                vpar_maxvmindig = None
                if self.maxvmindig and len(self.maxvmindig) >= i:
                    vpar_maxvmindig = self.maxvmindig[i-1]                
                vpar_vmiss = None
                if self.vmiss and len(self.vmiss) >= i:
                    vpar_vmiss = self.vmiss[i-1]                
                
                if not vpar_minvmiss or \
                   (maxval is not None and \
                    maxval >= type(maxval)(self.gen_vmiss(
                        type(maxval), mindig, maxdig, maxprec, 0, False))):
                    # add one order of magnitude (standard VMISS)
                    # exceptions are the x value or if minvmiss is set and
                    # the respective missing value does not appear in the data)
                    maxdig += 1
                    # correct minvmiss if maxval would be equal missing value:
                    vpar_minvmiss = False
                if vpar_minvmindig is not None and \
                   mindig < vpar_minvmindig:
                    mindig = vpar_minvmindig
                if vpar_maxvmindig is not None and \
                   mindig > vpar_maxvmindig:
                    mindig = vpar_maxvmindig
                force_scientific = None
                # a given vmiss from the caller can only extend the format:
                if vpar_vmiss:
                    if vrepr is None and vpar_vmiss.startswith('0x'):
                        # all values are None, miss is set to hex:
                        # change output format to hex
                        vrepr = HexInt
                    if vrepr == HexInt:
                        if not re.match(r'^0x(f+)$', vpar_vmiss, re.IGNORECASE):
                            raise ValueError(
                                "illegal vmiss for HexInt: {}".format(
                                    vpar_vmiss))
                        # the provided vmiss is HexInt and we have a HexInt
                        # series
                        (vmiss_mindig, vmiss_maxdig) = \
                            digits(HexInt(vpar_vmiss))
                        if vmiss_maxdig >= maxdig:
                            # use the provided vmiss value
                            maxdig = vmiss_maxdig
                            # mindig is alays 1 for HexInt
                    else:                        
                        reg = re.match(r"^([9]*)(\.([9]*))?([Ee](\+99))?$",
                                       self.vmiss[i-1])
                            # exponent in scientific notation must be +99
                            # there is no way to specify the number of digits in
                            # exponent in a format string (see gen_vformay), so
                            # it makes no sense to allow missing values other
                            # than E+99
                        if not reg:
                            raise ValueError(
                                "illegal vmiss for float/Decimal: {}".format(
                                    vpar_vmiss))
                        (vmiss_mindig, vmiss_maxdig) = \
                            digits(Decimal(vpar_vmiss))
                        if reg.group(4):
                            # scientific format missing value
                            if len(reg.group(1)) != 1:
                                raise ValueError(
                                    "vmiss in scientific notation must use one "
                                    "digit before comma: {}".format(
                                        vpar_vmiss))
                            # vmiss_maxdig is the highest possible digit value
                            # for the pattern. must not bee smaller than the
                            # highest digit value in the values.
                            # additionally the precision must be higher or equal
                            if maxdig-mindig <= vmiss_maxdig-vmiss_mindig and \
                                    maxdig < vmiss_maxdig:
                                # use the specified vmiss
                                mindig = vmiss_mindig
                                maxdig = vmiss_maxdig
                                maxprec = vmiss_maxdig - vmiss_mindig
                                if (vmiss_maxdig > 0 and vmiss_mindig > 0) or \
                                   (vmiss_maxdig < 0 and vmiss_mindig < 0):
                                    maxprec += 1
                            # else: don't use the actual vmiss, but still force
                            # scientific notation
                            force_scientific = True  # force scientific frmt
                        else:
                            # decimal format
                            if ((vmiss_mindig <= mindig or  # normal case
                                 (vmiss_mindig == 1 and mindig == 0))
                                 # special case mindig==0 if all None or all 0
                                and vmiss_maxdig >= maxdig):
                                # use the provided vmiss value
                                mindig = vmiss_mindig
                                maxdig = vmiss_maxdig
                            # else: don't use the actual vmiss, but still force
                            # decimal notation
                            force_scientific = False  # force decimal format
                self.data.VREPR.append(vrepr)
                self.data.VMISS.append(self.gen_vmiss(
                    vrepr, mindig, maxdig, maxprec, minval, vpar_minvmiss,
                    force_scientific=force_scientific))
                self.data.VFORMAT.append(self.gen_vformat(
                    vrepr, mindig, maxdig, maxprec, minval, vpar_minvmiss,
                    force_scientific=force_scientific))

    @staticmethod
    def gen_vmiss(vrepr, mindig, maxdig, maxprec, minval, minvmiss,
                  force_scientific=None):
        """
        Generates vmiss string for one variable.
        Parameters:
            vrepr         type of data variable (HexInt, float, Decimal)
            mindig        minimum digit used in the data
            maxdig        maximum digit used in the data
            maxprec       maximum precision used in a single value (important
                          for scientific notation)
            minval        minimum value used (important for accommodating a
                          minus sign)
            minvmiss      bool: uses minimal vmiss values (spare one digit)
                          in this case, the decimal format must be one char
                          wider (to accommodate a negative sign, which is else
                          wise accommodated by the longer pattern)
            force_scientific
                          True (force), False (force decimal),
                          None (do automatic)
        """
        if vrepr == HexInt:
            return "{}0x{}".format(
                " " if minval is not None and minval < 0 else "",
                "f" * maxdig)
        elif force_scientific is True or \
             (force_scientific is not False and
              NasaAmes1001Write._prefer_scientific(mindig, maxdig, maxprec)):
            return "{}9{}{}E+99".format(
                " " if minval is not None and minval < 0 else "",
                "." if maxprec > 1 else "",
                "9"*(maxprec-1))
        else:
            # decimal output format
            if minvmiss and minval is not None and minval < 0 and \
                    digits(minval)[1] == maxdig:
                # if minimal vmiss should be used and a negative minimum
                # value uses the maximum digit, we need to add one character
                # for the negative sign
                vmiss = " "
            else:
                vmiss = ""
            if maxdig > 0:
                vmiss += "9"*(maxdig)
            else:
                vmiss = "9"
            if mindig < 0:
                vmiss += "." + "9"*(-1*mindig)
            return vmiss

    @staticmethod
    def gen_vformat(vrepr, mindig, maxdig, maxprec, minval, minvmiss,
                    force_scientific=None):
        """
        Generates format string for one variable.
        Parameters:
            vrepr         type of data variable (HexInt, float, Decimal)
            mindig        minimum digit used in the data
            maxdig        maximum digit used in the data
            maxprec       maximum precision used in a single value (important
                          for scientific notation)
            minval        minimum value used (important for accommodating a
                          minus sign)
            minvmiss      bool: uses minimal vmiss values (spare one digit)
                          in this case, the decimal format must be one char
                          wider (to accommodate a negative sign, which is else
                          wise accommodated by the longer pattern)
            force_scientific
                          True (force), False (force decimal),
                          None (do automatic)
        """
        maxprec = max(maxprec, 1)
        if vrepr == HexInt:
            return "{}#0{:d}x".format(
                " " if minval is not None and minval < 0 else "",
                maxdig + (3 if minval is not None and minval < 0 else 2))
        if force_scientific is True or \
             (force_scientific is not False and
              NasaAmes1001Write._prefer_scientific(mindig, maxdig, maxprec)):
            width = maxprec+6 if minval is not None and minval < 0 \
                else maxprec+5
            if maxprec == 1:
                # format will be .0 --> no comma, one less wide
                width -= 1
            return "{:d}.{:d}E".format(width, maxprec-1)
        # decimal format
        maxdig = max(maxdig, 1)
        mindig = min(mindig, 0)
        if minvmiss and minval is not None and minval < 0 and \
                digits(minval)[1] == maxdig:
            # if minimal vmiss should be used and a negative minimum
            # value uses the maximum digit, we need to add one character
            # for the negative sign
            maxdig += 1
        if mindig < 0:
            return "{:d}.{:d}f".format(maxdig-mindig+1, -1*mindig)
        return "{0:d}.0f".format(maxdig)

    @staticmethod
    def _prefer_scientific(mindig, maxdig, maxprec):
        """
        Decides whether scientific representation should be preferred for
        for float/Decimal output.
        Parameters:
            maxdig   maximum digit used in variable (order of magnitude)
            mindig   minimum digit used in variable (order of magnitude)
            maxprec  maximum precision used in one value
        """
        # calculate the deciamls used in decimal notations:
        if mindig < 0:
            decused = max(maxdig, 1)-mindig
        else:
            decused = maxdig - min(mindig, 1)
        if (maxdig < -3 or mindig > 3 or decused > 7) and \
            decused-4 > maxprec:
            # if numbers are more than 3 orders of magnitude off E00 or
            # the range spans more than 9 orders of magnitudes, it's
            # reasonable to use scientifig format.
            # but only if scientific mantissa (maxprec) is reasonably shorter
            # then the decimal representation (basically, it must compensate for
            # additional "E..."
            return True
        return False

    def _write(self):
        """
        Writes the file to opened file handle.
        This method is only needed to extract the write functionality in order
        to wrap the 'with open(filename)' around it in the read() method.
        Parameters:
            None
        Returns:
            None
        Raises:
            NasaAmesError     on any unrecoverable read error
        """
        for self.state in range(len(self.__class__.SECTIONS)):
            statename = self.curr_state_name()
            try:
                method = getattr(self, '_write_' + statename.lower())
            except AttributeError:
                # If no dedicated _write method exists: write one line as
                # string.
                # Those are: ONAME, ORG, SNAME, MNAME, XNAME
                self.logger.debug(u'write {}: {}'.format(statename,
                                                         self.data[statename]))
                self.writeline(self.data[statename])
            else:
                method()

    def _write_nlhead(self):
        """
        Writes the NLHEAD line to file.
        """
        self.writeline('{0:d} 1001'.format(self.data.NLHEAD))


    def _write_ivol_nvol(self):
        """
        Writes the IVOL/NVOL line to file.
        """
        self.writeline('{0:d} {1:d}'.format(self.data.IVOL, self.data.NVOL))

    def _write_date_rdate(self):
        """
        Writes the DATE/RDATE line to file.
        """
        self.writeline(self.data.DATE.strftime("%Y %m %d ") +\
                       self.data.RDATE.strftime("%Y %m %d"))

    def _write_dx(self):
        """
        Writes DX to the file.
        """
        if self.data.DX == 0.0:
            self.writeline('0')
        elif self.xmindig != None and self.xmindig < 0:
            self.writeline('{:.{}f}'.format(self.data.DX,
                                            self.xmindig * -1))
        elif self.xmindig != None and self.xmindig >= 0:
            self.writeline('{:.0f}'.format(self.data.DX))
        else:
            self.writeline('{:f}'.format(self.data.DX))

    def _write_vscal(self):
        """
        Writes the VSCAL line to file.
        """
        self.writeline(" ".join([str(vscal) for vscal in self.data.VSCAL]))

    def _write_vmiss(self):
        """
        Writes the VMISS line to file.
        """
        self.writeline(" ".join([x.lstrip() for x in self.data.VMISS]))

    def _write_vname(self):
        """
        Writes the VMISS line to file.
        """
        for vname in self.data.VNAME:
            self.writeline(vname)

    def _write_scoml(self):
        """
        Writes the VMISS line to file.
        """
        self.writeline(len(self.data.SCOML))
        for scom in self.data.SCOML:
            self.writeline(scom)

    def _write_ncoml(self):
        """
        Writes the VMISS line to file.
        """
        self.writeline(len(self.data.NCOML))
        for ncom in self.data.NCOML:
            self.writeline(ncom)

    def _write_data(self):
        """
        Writes data lines to files.
        """
        if len(self.data.DATA) == 0:
            return
        fmt = ['{{:{}}}'.format(self.xformat)]
        miss = [None]
        for j in range(self.data.NV):
            fmt.append("{{0:{0}}}".format(self.data.VFORMAT[j]))
            miss.append(self.data.VMISS[j])

        for i in range(len(self.data.DATA[0])):
            line = fmt[0].format(self.data.DATA[0][i])
            for j in range(1, len(self.data.DATA)):
                if self.data.DATA[j][i] is None:
                    line += ' ' + miss[j]
                else:
                    line += ' ' + fmt[j].format(self.data.DATA[j][i])
            self.writeline(line)

    def writeline(self, line):
        """
        Writes one line to the file and increases the objects lnum attribute.
        Parameters:
            line     line of text to write
        Returns:
            None
        """
        self.lnum += 1
        if line is None:
            line = ''
        elif not isinstance(line, string_types):
            line = unicode(line) if PY2 else str(line)
        line += '\n'
        if PY2:
            self.file.write(line.encode(self.encoding))
        else:
            # still some encoding chaos...
            # later we should just use a text mode for the file.
            # after dropping support for py2
            if isinstance(self.file, io.BufferedIOBase):
                self.file.write(line.encode(self.encoding))
            else:
                self.file.write(line)
        self.logger.debug(u"write line {}: {}".format(self.lnum, line))



"""
ebas/nasa_ames/read/parse_variables.py
$Id: parse_variables.py 2734 2021-11-17 17:01:45Z pe $

parser for file input functionality for EBAS NASA Ames module
parses variables from NasaAmes1001 object and builds NasaAmes

History:
V.1.0.0  2013-06-22  pe  initial version

"""

import re
from nilutility.datatypes import DataObject
from nilutility.string_helper import list_splitter
from .parse_ebasmetadata import NasaAmesPartialReadParserEbasMetadata
from .legacy_vnames import NasaAmesPartialReadLegacyVnames

class NasaAmesPartialReadParserVariables(# pylint: disable=R0901, W0223,
        NasaAmesPartialReadLegacyVnames,
        NasaAmesPartialReadParserEbasMetadata):
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    # W0223: ... is abstract in base class but is not overridden
    # this class is abstract as well (does not have any NotImplementedError)
    """
    Parser for Nasa Ames I/O object.
    This is a base class for NasaAmesPartialRead
    """
    def _parse_variables(self, skip_variables=None, ignore_parameter=False):
        """
        Parse VNAME lines (variable names).
        Additionally checks the missing values for each variable.
        Syntax:
            "param_name, unit[,Keyword=Value,...]"
        Parameters:
            skip_variables    list of variable numbers to be skipped (variabzle
                              numbers start with 1 for the first variable after
                              end_time)
            ignore_parameters ignore errors related to paramters and units
                              this is needed when a file with non standard
                              vnames should be processed without importing it
                              into the ebas.domain.
        Returns:
            None
        """
        self._skip_variables(skip_variables)
        self.translate_legacy_vnames()
        self.parse_variable_endtime()

        i = 0
        for i in range(1, self.nasa1001.data.NV):
            if i in self.internal.read_tmp.var_skipped:
                continue
            vname = self.nasa1001.data.VNAME[i]
            vmiss = self.nasa1001.data.VMISS[i]
            if vname.startswith("numflag"):
                self._parse_variable_numflag(i, vname, vmiss)
            else:
                self._parse_variable_regular(i, vname, vmiss,
                                             ignore_parameter=ignore_parameter)
        # check if there's at least one variable
        if len(self.variables) < 1 and self.internal.read_tmp.var_skipped:
            self.warning('No variable found (all skipped)', lnum=13+i)
        elif len(self.variables) < 1:
            self.error('No variable found', lnum=13+i)
        # check if last variable has a flag
        elif not self.variables[-1].flagcol:
            self.error('Last variabe must have a flag', lnum=13+i)
            # This is a critical error, checking data crashes with empty flags!
            self.internal.read_tmp.hea_error = True

    def _skip_variables(self, skip_variables):
        """"
        Skip variables by caller.
        Parameters:
            skip_variables    list of variable numbers to be skipped (variabzle
                              numbers start with 1 for the first variable after
                              end_time)
        Returns:
            None
        """
        if skip_variables is None:
            return
        for i in skip_variables:
            if i < 1 or i >= self.nasa1001.data.NV:
                self.error("skip_variables: variable number {} is illegal "
                           "(must be between 1 and NV-1 (1..{} in this case))"
                           .format(i, self.nasa1001.data.NV-1))
            elif self.nasa1001.VNAME[i].startswith("numflag"):
                self.error("skip_variables: variable number {} is illegal "
                           "(flag variable may not be skipped)".format(i))
            else:
                self.internal.read_tmp.var_skipped.append(i)
                self.warning(
                    "VNAME[{}]: variable skipped: {}".format(
                        i, self.nasa1001.VNAME[i]),
                    lnum=13+i)

    def parse_variable_endtime(self):
        """
        Parse VNAME line for end_time variable (first dep. variable).
        Parameters:
            None
        Returns:
            None
        """
        vname = self.nasa1001.data.VNAME[0]
        if vname not in \
           ("end_time of measurement, days from the file reference point",
            "end_time, Julian date counted from the file reference point",
            "end_time, days from the file reference point"):
            self.error(
                "VNAME[0] is not a supported end_time definition: '{}', should "
                "be 'end_time, days from the file reference point'".
                format(vname), lnum=13)
            self.internal.read_tmp.hea_error = True
            return   # stop parsing
        if not re.match(r'^9+(\.9*)?$', self.nasa1001.data.VMISS[0]) and \
           not re.match(r'^9(\.9*)?E\+?9*$', self.nasa1001.data.VMISS[0], re.I):
            self.error(
                "VMISS[0]: format error: '{}'. Should be '9.9*E+99', '9*.9*' "
                "or similar. Variable skipped.".format(
                    self.nasa1001.data.VMISS[0]),
                lnum=12)
            self.internal.read_tmp.hea_error = True
            return   # stop parsing

    @staticmethod
    def _is_general_flagcolumn(vname):
        """
        Check if vname is a general flag column (i.e. no comp name)
        Parameters:
            vname    vname line of the flag column
        Returns:
            True/False
        """
        if vname == "numflag" or \
           vname == "numflag, no unit":
            return True
        return False

    def _is_dedicated_flagcolumn(self, vname):
        """
        Check if vname is a dedicated flag column (i.e. includes comp name)
        Parameters:
            vname    vname line of the flag column
        Returns:
            True/False
        """
        if vname.startswith("numflag " + \
                            self.variables[-1].metadata.comp_name + \
                            ", no unit"):
            return True
        return False

    def _parse_variable_numflag(self, vnum, vname, vmiss):
        """
        Parse VNAME line for a flag column.
        Parameters:
            vnum    variable number (in the file)
            vname   vname line read from file (str)
            vmiss   vmiss definition (str)
        Returns:
            None
        Variable number in the object differs from variable number in file
        (because the object only stores measurements as variables, having
        flags as an attribute)
        """
        if vnum-1 in self.internal.read_tmp.var_skipped and \
           (len(self.variables) == 0 or self.variables[-1].flagcol):
            # if prev variable has been skipped, the flag column must also be
            # skipped. there is no element for the variable in the list!
            self.warning(
                "VNAME[{}]: flag skipped (flag for skipped variable)".
                format(vnum), 13 + vnum)
            self.internal.read_tmp.var_skipped.append(vnum)
            return
        if len(self.variables) == 0 or \
           self.variables[-1].flagcol:
            self.error(
                'VNAME[{}]: numflag without preceding variable, flag skipped'.
                format(vnum), lnum=13+vnum)
            self.internal.read_tmp.var_skipped.append(vnum)
            return  # stop parsing

        if self.nasa1001.data.NV == 3:
            # this may be a general or a dedicated flag column
            if not self._is_general_flagcolumn(vname) and \
               not self._is_dedicated_flagcolumn(vname):
                self.warning(
                    u"VNAME[{}]: unconventional VNAME for flag column: '{}'. "
                    u"Should be 'numflag, no unit' or start with 'numflag {}, "
                    u"no unit'".format(vnum, vname,
                                       self.variables[-1].metadata.comp_name),
                    lnum=13+vnum)
        elif len(self.variables) > 1 and not self.variables[-2].flagcol:
            # this must be a general flag column
            if not self._is_general_flagcolumn(vname):
                self.warning(
                    u"VNAME[{}]: unconventional VNAME for flag column: '{}'. "
                    u"Should be 'numflag, no unit'".format(vnum, vname),
                    lnum=13+vnum)
        else:
            # this must be a dedicated flag column
            if not self._is_dedicated_flagcolumn(vname):
                self.warning(
                    u"VNAME[{}]: unconventional VNAME for flag column: '{}'. "
                    u"Should start with 'numflag {}, no unit'".format(
                        vnum, vname, self.variables[-1].metadata.comp_name),
                    lnum=13+vnum)

        self.variables[-1].flagcol = True
        # parse vmiss for this flag column
        reg = re.match(r"^9\.(9+)$", vmiss)
        if reg:
            flen = len(reg.group(1))
            if flen % 3 != 0:
                self.error(
                    "VMISS[{}]: illegal missing value for flag column: '{}'. "
                    "Should be 9.[999]... (groups of 3 digiits). Flag skipped".
                    format(vnum, vmiss), lnum=12)
                self._skip_last_flag(vnum)
        else:
            self.error(
                "VMISS[{}]: illegal missing value for flag column: '{}'. "
                "Should be 9.[999]... (groups of 3 digiits). Flag skipped".
                format(vnum, vmiss), lnum=12)
            self._skip_last_flag(vnum)

    def _skip_last_flag(self, vnum):
        """
        Skip a flag column (due to errors in the column definition)
        Skipped column is the flag column for the last variable in the object,
        the column number in the file is vnum.
        ParametersL
            vnum    variable number (in the file)
        Returns:
            None:
        """
        self.internal.read_tmp.var_skipped.append(vnum)  # skip flag var
        # additionally skip already existing data variables which
        # use this flag column
        skip = []
        self.variables[-1].flagcol = False
        for i in reversed(range(0, len(self.variables))):
            if self.variables[i].flagcol:
                break
            skip.append(self.internal.read_tmp.var_num[i])
            self._skip_existing_variable(i)
        self.error("skipped variables {} due to skipped flag variable {}".
                   format(", ".join([str(x) for x in sorted(skip)]), vnum),
                   lnum=12)


    def _skip_existing_variable(self, vnum):
        """
        Skips an already existing variables.
        Parameters:
            vnum   variable number (in i/o object)
        Returns:
            None
        """
        j = self.internal.read_tmp.var_num[vnum]  # pos in file
        self.internal.read_tmp.var_skipped.append(j)
        del self.internal.read_tmp.var_num[vnum]
        del self.internal.read_tmp.vmiss[vnum]
        del self.variables[vnum]


    def _parse_variable_regular(self, vnum, vname, vmiss,
                                ignore_parameter=False):
        """
        Parse VNAME line and VMISS specification for a regular variable.
        Parameters:
            vnum    variable number (in the file, 1=first var after endtime)
            vname   vname line read from file (str)
            vmiss   vmiss definition (str)
           ignore_parameters
                    ignore errors related to paramters and units this is needed
                    when a file with non standard vnames should be processed
                    without importing it into the ebas.domain.
        Returns:
            None
        Variable number in the object differs from variable number in file
        (because the object only stores measurements as variables, having
        flags as an attribute)
        """
        if not re.match(r'^9+(\.9*)?$', vmiss) and \
           not re.match(r'^9(\.9*)?E\+?9*$', vmiss, re.I) and \
           not re.match(r'^0x9+$', vmiss, re.I) and \
           not re.match(r'^0xF+$', vmiss, re.I):
            self.error(
                "VMISS[{}]: format error: '{}'. Should be '9.9*E+99', '9*.9*', "
                "'0x9*', '0xF*' or similar. Variable skipped."
                .format(vnum, vmiss), lnum=12)
            self.internal.read_tmp.var_skipped.append(vnum)
            return   # stop parsing
        if vmiss.lower().startswith('0x'):
            is_hex = True
        else:
            is_hex = False

        reg = re.match(r"^([^,]+),\s*([^,]+)(,\s*(.*))?$", vname)
        if not reg:
            self.error(
                "VNAME[{}]: illegal syntax for VNAME: '{}'. Should be "
                "'component_name, unit[, metadata=value]...'. Variable skipped"
                .format(vnum, vname), lnum=13+vnum)
            self.internal.read_tmp.var_skipped.append(vnum)
            return # stop parsing this line
        comp_name, unit = self._parse_ebasmetadata_vname_positional(
            reg.group(1).strip(), reg.group(2).strip(), 13+vnum, vnum,
            ignore_parameter=ignore_parameter)

        # append the variable:
        self.variables.append(
            DataObject(is_hex=is_hex,
                       values_=[], flags=[], flagcol=False,
                       metadata=DataObject(comp_name=comp_name,
                                           unit=unit)))
        self.internal.read_tmp.var_num.append(vnum)
        self.internal.read_tmp.vmiss.append(vmiss)

        if self.metadata.datadef == 'EBAS_1' and reg.group(4):
            self.warning(
                "VNAME[{}]: additional metadata in VNAME in EBAS_1".format(
                    vnum), lnum=13+vnum)

        def _resort_meta(metalist):
            """
            Special: instrument type must be processed first (some other
            parsers depend on it).
            """
            int_list = list(metalist)
            for elem in int_list:
                if elem.startswith('Instrument type'):
                    yield elem
            for elem in int_list:
                if not elem.startswith('Instrument type'):
                    yield elem

        # parse additional metadata (variable specific metadata, statistics,
        # characteristics)
        if reg.group(4):
            for add in _resort_meta(list_splitter(reg.group(4), ",")):
                try:
                    (tag, value) = list_splitter(add, "=")
                except ValueError:
                    self.error(
                        "VNAME[{}]: additional metadata element '{}' not in "
                        "'tag=value' syntax".format(vnum, add),
                        lnum=13+vnum)
                    continue
                tag = tag.strip()
                value = value.strip()
                if value == u'':
                    value = None
                var_index = len(self.variables)-1
                self._parse_ebasmetadata(tag, value, 13 + vnum,
                                         var_index=var_index)



"""
Module atmos_phys.convert.base: Base clas for all unit conversions.
"""

from nilutility import numeric


class ConvertBase(object):
    """Base class for conversions.

    Attributes:
        from_unit: source unit of conversion
        to_unit: target unit for conversion
        roundoffset: rounding offset
        maxround: maximum rounding threshold
        rounding: actual rounding digits used
        parameters: additional parameters for the conversion
    """

    def __init__(self, from_unit, to_unit, roundoffset=0, maxround=None):
        """
        Set up conversion object.
        Parameters:
            from_unit:
            to_unit: conversion from and to unit. Must be checked in
                     _set_parameters of the final class.
            roundoffset: rounding offset (usually +1 for import and -1 for
                         export)
            maxround: maximum rounding threshold, do no round coarser then
                      this should be used for export conversion only
                      (rounding expressed like in round function, 0=integer,
                      1=1digit after comma).
                      Background: Historical data for some components have a
                      rounding issue: e.g. ozone, historical data have
                      been converted on user side, but reported only in
                      integer precission ug/m3. Now the exporter converts
                      back to nmol/mol, but dropps one decimal in rounding,
                      resulting in nmol/mol rounded to 10.
                      In order to avoid this, it's possible to specify a
                      threshold for rounding on export depending on
                      component.
                      Use an integer acording to the round function
                      (0=rounded to integer 1=round 1 decimal after the
                      comma, etc.). Alternatively specify None if no
                      threshold is needed.
        """
        self.from_unit = from_unit
        self.to_unit = to_unit
        self.roundoffset = roundoffset
        self.maxround = maxround
        self.rounding = None
        self.parameters = None
        self._set_parameters()

    def _set_parameters(self):
        """
        Checks frokm and to unit. Sets the conversion parameters.
        Must be implemented in derived class.
        Parameters:
            None
        Returns:
            None
        """
        raise NotImplementedError()

    def conversion_string(self, from_val=None, to_val=None):
        """
        Generates a conversion parameter string for the conversion.
        E.g. "from nmol/mol to ug/m3 at standard conditions (273.15 K, 1013.25 "
                 "hPa), using factor 1.234.",
             "from nmol/mol to ug/m3 at standard conditions (273.15 K, 1013.25 "
                 "hPa), no conversion factor used (no nonzero, valid values)."
        Parameters:
            from_val:
            to_val:
                If from_val and to_val are given, the values are included in the
                string (used to document conversion of uncertainty or detection
                limit values).
        Returns:
            a string for documenting the conversion.

        Must be implemented in derived class.
        """
        raise NotImplementedError()

    def _calc_base_rounding(self, values, add_offset=0):
        """
        Calculate the right number of digits for rounding according to the
        actual values' precision.
        Default implementation (without taking into account effects of e.g.
        conversion factors. See _calc_rounding and the use of add_offset.
        Parameters:
            values: list of data values to be converted
            add_offset: additional rounding offset for effects not taken into
                        account in this default implementation (e.g. conversion
                        factor)
        Returns:
            rounding needed (digits relative to the comma)
            e.g. 0 --> rounding to integer, 1 --> rounding to 10s,
            -1 --> rounding to 1 dicit after comma

            None in case of all values None
        """
        digstat = numeric.digits_stat(values)
        if digstat[2] == 0:
            # special case, all values were 0.0: double check and return
            if [x for x in values if x != 0.0 and x is not None] != []:
                raise RuntimeError("digits_stat returned unexpected value")
        if digstat[2] is None:
            # special case, all values were None: double check and return
            if [x for x in values if x is not None] != []:
                raise RuntimeError("digits_stat returned unexpected value")
            return None

        if digstat[2] > 0:
            rnd = digstat[2] * -1 + 1
        else:  # < 0
            rnd = digstat[2] * -1

        # add additinal offset depending on client class:
        rnd += add_offset

        # add or subtract on digit for rounding as specified by caller:
        rnd += self.roundoffset

        if self.maxround is not None and rnd < self.maxround:
            rnd = self.maxround
        return rnd

    def _calc_rounding(self, values):
        """
        Calculate the number of digits for rounding. This method is meant to be
        overridden by client implementations which need to take care of
        additional effects (e.g. conversion factor).
        """
        # Default, add_offset=0; override this method if needed
        return self._calc_base_rounding(values, add_offset=0)

    def get_rounding(self, values):
        """
        Calculate and store the right number of digits for rounding according
        to the actual values' precision.
        """
        if self.rounding is None:
            self.rounding = self._calc_rounding(values)
        return self.rounding

    def convert_data(self, values):
        """
        Convert values using the conversion factor.
        Parameters:
            values: list of values (list of float)
        Returns:
            None
        Must be implemented in derived class.
        """
        raise NotImplementedError()



class ConvertBaseFactor(ConvertBase):  # pylint: disable=W0223
    # W0223: Method ... is abstract in base class but is not overridden
    """
    Base class for conversions with only a factor.
    """

    def conversion_string(self, from_val=None, to_val=None):
        """
        Generates a conversion parameter string for the conversion.
        E.g. "from Torr to hPa, using factor 1.33322",
             "from Torr to hPa, no conversion factor used (no nonzero, "
             "valid values)"
        Parameters:
            from_val:
            to_val:
                If from_val and to_val are given, the values are included in the
                string (used to document conversion of uncertainty or detection
                limit values).
        Returns:
            a string for documenting the conversion.
        """
        from_val = "{} ".format(from_val) if from_val is not None else ""
        to_val = "{} ".format(to_val) if to_val is not None else ""
        if self.rounding is None:
            convf = "no conversion factor (no nonzero, valid values)"
        else:
            convf = "conversion factor {:.{rdigits}f}"\
                    .format(self.parameters.factor, rdigits=self.rounding+2)
        return ("from '{}{}' to '{}{}', {}").format(
            from_val, self.from_unit, to_val, self.to_unit, convf)

    def _calc_rounding(self, values):
        """
        Calculate and store the right number of digits for rounding according
        to the actual values' precision.
        """
        # add order of magnitude for conversion factor:
        offs = 0
        (mant, ordm) = numeric.frexp10(self.parameters.factor)
        if mant >= 5:
            ordm += 1
        offs -= ordm
        return super(ConvertBaseFactor, self)._calc_base_rounding(
            values, add_offset=offs)

    def convert_data(self, values):
        """
        Convert values using the conversion factor.
        Parameters:
            values: list of values (list of float)
        Returns:
            None
        """
        rnd = self.get_rounding(values)

        for i, val in enumerate(values):
            if val is not None:
                values[i] = \
                    round(val*self.parameters.factor, rnd)



"""
ebas.domain.basic_domain_logic.value_check
$Id: value_check.py 2780 2022-03-09 13:20:02Z pe $
Basic ebas domain logic for value checks (boundary and spike checks).
"""

from nilutility.statistics import running_median, running_stddev
from .. import EbasDomainError
from .flags import get_flag_summary
from .base import EbasDomainIssueList, \
    ISSUE_SEVERITY_ERROR, ISSUE_SEVERITY_WARNING, \
    MSGID_VALUECHECK_BNDCK_LOWER, MSGID_VALUECHECK_BNDCK_UPPER, \
    MSGID_VALUECHECK_SPKCK_LOWER, \
    MSGID_VALUECHECK_SPKCK_UPPER, MSGID_VALUECHECK_SPKCK_STD_LOWER, \
    MSGID_VALUECHECK_SPKCK_STD_UPPER

class _BaseEbasDomBNDIssue(EbasDomainError):
    """
    Base class for boundary/spike check inconsistency exceptions.
    """
    def __init__(self, issues):
        """
        Parameters:
            issues   list of issues, each a tuple (severity, msg-id, value#,
                     message)
        """
        str_ = str([tuple(isu[key] for key in issues.keys)
                    for isu in issues])
        EbasDomainError.__init__(self, str_)
        self.issues = issues

class EbasDomBNDError(_BaseEbasDomBNDIssue):
    """
    Exception class for boundary/spike check inconsistencies (checks for lists).
    """
    def __init__(self, issues):
        """
        Parameters:
            issues   list of issues, each a tuple (severity, msg-id, value#,
                     message)
        """
        _BaseEbasDomBNDIssue.__init__(self, issues)

class EbasDomBNDWarning(_BaseEbasDomBNDIssue):
    """
    Exception class for boundary/spike check inconsistencies (checks for lists).
    """
    def __init__(self, issues):
        """
        Parameters:
            issues   list of issues, each a tuple (severity, msg-id, value#,
                     message)
        """
        _BaseEbasDomBNDIssue.__init__(self, issues)

def check_values(proj, values, flags,
                 minval, maxval,
                 spike_radius,
                 spike_minoffs, spike_maxoffs,
                 spike_stddev_minfact, spike_stddev_maxfact,
                 fix_bc211=False):
    """
    Perform all value checks of data series (values, flags). This includes
    a boudary check and spike detection.
    Parameters:
        proj          project which defines the rule (None or project acronym)
        values        list of values
        flags         list of flags (list of tuples)
                      !! be sure to pass the original list, flags may be
                      !! appenden when using fix_bc211
        minval        minimum allowed (absolute) value
        maxval        minimum allowed (absolute) value
        spike_radius  radius for spike detection
        spike_minoffs minimum factor for deviation from running median
        spike_maxoffs maximum factor for deviation from running median
        spike_stddev_minfact
                      minimum factor of stddev deviation from running median
        spike_stddev_maxfact
                      maximum factor of stddev deviation from running median
        fix_bc211     flag boundary check issues with flag 211
    Returns:
        EbasDomainIssueList with issues found (empty list if OK)
    """
    return EbasDomainIssueList(sorted(
        boundary_check(
            proj, values, flags, minval, maxval, fix_bc211=fix_bc211) +
        spike_check(
            proj, values, flags, spike_radius,
            spike_minoffs, spike_maxoffs, spike_stddev_minfact,
            spike_stddev_maxfact,
            fix_bc211=fix_bc211),
        key=lambda x: (x['row'], x['msg_id'])))

def boundary_check(proj, values, flags, minval, maxval, fix_bc211=False):
    """
    Boundary check of data series (values, flags).
    Parameters:
        proj          project which defines the rule (None or project acronym)
        values        list of values
        flags         list of flags (list of tuples)
                      !! be sure to pass the original list, flags may be
                      !! appenden when using fix_bc211
        minval        minimum allowed (absolute) value
        maxval        minimum allowed (absolute) value
        fix_bc211     flag boundary check issues with flag 211
    Returns:
        EbasDomainIssueList with issues found (empty list if OK)
    """
    proj = " ({})".format(proj) if proj else ""
    issues = EbasDomainIssueList()
    if minval is None and maxval is None:
        # no boundaries defined, no issues to report
        return issues
    for i, value in enumerate(values):
        if value is None:
            continue
        if minval is not None and value < minval and \
                not bc_exceptions(flags[i], False):
            if fix_bc211:
                flags[i].append(211)
                issues.add(
                    ISSUE_SEVERITY_WARNING, MSGID_VALUECHECK_BNDCK_LOWER, i,
                    "Boundary check{}: value {} is less than defined "
                    "minimum {}, --fix bc211: adding flag 211".format(
                        proj, value, minval))
            else:
                issues.add(
                    ISSUE_SEVERITY_ERROR, MSGID_VALUECHECK_BNDCK_LOWER, i,
                    "Boundary check{}: value {} is less than defined "
                    "minimum {}".format(proj, value, minval))
        if maxval is not None and value > maxval and \
                not bc_exceptions(flags[i], True):
            if fix_bc211:
                flags[i].append(211)
                issues.add(
                    ISSUE_SEVERITY_WARNING, MSGID_VALUECHECK_BNDCK_UPPER, i,
                    "Boundary check{}: value {} is greater than defined "
                    "maximum {}, --fix bc211: adding flag 211".format(
                        proj, value, maxval))
            else:
                issues.add(
                    ISSUE_SEVERITY_ERROR, MSGID_VALUECHECK_BNDCK_UPPER, i,
                    "Boundary check{}: value {} is greater than defined "
                    "maximum {}".format(proj, value, maxval))
    return issues

def spike_check(
        proj, values, flags, spike_radius, spike_minoffs, spike_maxoffs,
        spike_stddev_minfact, spike_stddev_maxfact, fix_bc211):
    """
    Spike check of data series (values, flags).
    Parameters:
        proj          project which defines the rule (None or project acronym)
        values        list of values
        flags         list of flags (list of tuples)
                      !! be sure to pass the original list, flags may be
                      !! appenden when using fix_bc211
        spike_radius  radius for spike detection
        spike_minoffs minimum factor for deviation from running median
        spike_maxoffs maximum factor for deviation from running median
        spike_stddev_minfact
                      minimum factor of stddev deviation from running median
        spike_stddev_maxfact
                      maximum factor of stddev deviation from running median
        fix_bc211     flag boundary check issues with flag 211
    Returns:
        EbasDomainIssueList with issues found (empty list if OK)
    """
    proj = " ({})".format(proj) if proj else ""
    issues = EbasDomainIssueList()
    if spike_radius is None and spike_minoffs is None and \
            spike_maxoffs is None and spike_stddev_minfact is None and \
            spike_stddev_maxfact is None:
        # no spike definitions, no issues to report
        return issues
    # modify values, only take those which are not flagged invalid
    values = [val if get_flag_summary(fla)[1] == 0 else None
              for val, fla in zip(values, flags)]
    medians = running_median(values, spike_radius)
    stddevs = []
    if spike_stddev_minfact is not None or spike_stddev_maxfact is not None:
        # we need to calculate running stddevs:
        stddevs = running_stddev(values, spike_radius)
    for i in range(len(values)):  # pylint: disable=C0200
        # C0200: Consider using enumerate instead of iterating with range/len
        # --> use same index for several lists, alternative would be a threefold
        # zip and additionally, I would need the index, which makes
        # an enumerate(zip(x, y, z)) -- ugly and less readable
        if values[i] is None:
            continue
        if spike_minoffs is not None and \
                values[i] < medians[i] + spike_minoffs and \
                not bc_exceptions(flags[i], False):
            if fix_bc211:
                flags[i].append(211)
                issues.add(
                    ISSUE_SEVERITY_WARNING, MSGID_VALUECHECK_SPKCK_LOWER, i,
                    "Spike check{}: value ({}) is less than running median "
                    "({}) {} {}; --fix bc211: adding flag 211".format(
                        proj, values[i], medians[i],
                        "-" if spike_minoffs < 0 else "+",
                        abs(spike_minoffs)))
            else:
                issues.add(
                    ISSUE_SEVERITY_ERROR, MSGID_VALUECHECK_SPKCK_LOWER, i,
                    "Spike check{}: value ({}) is less than running median "
                    "({}) {} {}".format(
                        proj, values[i], medians[i],
                        "-" if spike_minoffs < 0 else "+",
                        abs(spike_minoffs)))
        if spike_maxoffs is not None and \
                values[i] > medians[i] + spike_maxoffs and \
                not bc_exceptions(flags[i], True):
            if fix_bc211:
                flags[i].append(211)
                issues.add(
                    ISSUE_SEVERITY_WARNING, MSGID_VALUECHECK_SPKCK_UPPER, i,
                    "Spike check{}: value ({}) is greater than running median "
                    "({}) {} {}; --fix bc211: adding flag 211".format(
                        proj, values[i], medians[i],
                        "-" if spike_maxoffs < 0 else "+",
                        abs(spike_maxoffs)))
            else:
                issues.add(
                    ISSUE_SEVERITY_ERROR, MSGID_VALUECHECK_SPKCK_UPPER, i,
                    "Spike check{}: value ({}) is greater than running median "
                    "({}) {} {}".format(
                        proj, values[i], medians[i],
                        "-" if spike_maxoffs < 0 else "+",
                        abs(spike_maxoffs)))
        if spike_stddev_minfact is not None and stddevs[i] is not None and \
                values[i] < medians[i] + spike_stddev_minfact * stddevs[i] and \
                not bc_exceptions(flags[i], False):
                # stddevs[i] is None when there is less then 2 valid values
            if fix_bc211:
                flags[i].append(211)
                issues.add(
                    ISSUE_SEVERITY_WARNING, MSGID_VALUECHECK_SPKCK_STD_LOWER, i,
                    "Spike check{}: value ({}) is less than running median "
                    "({}) {} {} * running stddev ({}); "
                    "--fix bc211: adding flag 211".format(
                        proj, values[i], medians[i],
                        "-" if spike_stddev_minfact < 0 else "+",
                        abs(spike_stddev_minfact), stddevs[i]))
            else:
                issues.add(
                    ISSUE_SEVERITY_ERROR, MSGID_VALUECHECK_SPKCK_STD_LOWER, i,
                    "Spike check{}: value ({}) is less than running median ({}) "
                    "{} {} * running stddev ({})".format(
                        proj, values[i], medians[i],
                        "-" if spike_stddev_minfact < 0 else "+",
                        abs(spike_stddev_minfact), stddevs[i]))
        if spike_stddev_maxfact is not None and stddevs[i] is not None and \
                values[i] > medians[i] + spike_stddev_maxfact * stddevs[i] and \
                not bc_exceptions(flags[i], True):
                # stddevs[i] is None when there is less then 2 valid values
            if fix_bc211:
                flags[i].append(211)
                issues.add(
                    ISSUE_SEVERITY_WARNING, MSGID_VALUECHECK_SPKCK_STD_UPPER, i,
                    "Spike check{}: value ({}) is greater than running median "
                    "({}) {} {} * running stddev ({}); "
                    "--fix bc211: adding flag 211".format(
                        proj, values[i], medians[i],
                        "-" if spike_stddev_maxfact < 0 else "+",
                        abs(spike_stddev_maxfact), stddevs[i]))
            else:
                issues.add(
                    ISSUE_SEVERITY_ERROR, MSGID_VALUECHECK_SPKCK_STD_UPPER, i,
                    "Spike check{}: value ({}) is greater than running median "
                    "({}) {} {} * running stddev ({})".format(
                        proj, values[i], medians[i],
                        "-" if spike_stddev_maxfact < 0 else "+",
                        abs(spike_stddev_maxfact), stddevs[i]))
    return issues

def bc_exceptions(flags, upper_bnd):
    """
    Check if an exception for the boundary checks should be made.
    There is a defined list of valid flags that make an exception. In
    addition, any invalid flag makes an exception for the boundary checks.
    Parameters:
        flags      list of flags for the sample
        upper_bnd  bool, should the upper (True) or lower (False) boudary
                   be checked
    """
    # some valid flags make an exception:
    flaglist_exception = [100, 111, 211]
    if upper_bnd:
        flaglist_exception += [110, 210, 410, 559]
    else:
        flaglist_exception += [147]
    if any([x in flaglist_exception for x in flags]):
        return True

    # all invald flags make an exception:
    if get_flag_summary(flags)[1] > 0:
        return True
    return False



"""
nilutility/lock.py
$Id: lock.py 1388 2016-09-15 09:05:32Z pe $

Simple lock file handling
"""

import os
import logging

class LockError(Exception):
    """
    Exception class, raised if something strange happened with locks.
    """
    pass

class Lock(object):
    """
    Implements a simple lock file mechanism.
    """

    def __init__(self, lockfile):
        """
        Initialize locking class
        """
        self.status = False
        self.lockfile = lockfile
        self.logger = logging.getLogger('lock')

    def lock(self):
        """
        Lock the lockfile.
        Parameters:
            None
        Retiurn:
            True when locked, False when locking could not be aquired.
        """
        if os.path.isfile(self.lockfile):
            self.logger.error('no lock aquired (lockfile exists: {})'
                .format(self.lockfile))
            return False
        open(self.lockfile, 'a').close()
        self.status = True
        self.logger.debug('lock aquired (lockfile created: {})'
            .format(self.lockfile))
        return True

    def release(self):
        if not self.status:
            self.logger.error('lock: called release without a lock')
            raise LockError("Release without a lock")
        if not os.path.isfile(self.lockfile):
            self.logger.error('lock file vanished: {}'
                .format(self.lockfile))
            raise LockError("Lockfile vanished")
        os.remove(self.lockfile)
        self.logger.debug('lock released (lockfile deletd: {})'
            .format(self.lockfile))
        self.status = False



"""
Fallback class to be used when ebas.domain is not available (e.g. EbasIO Module
published for 3rd parties).

$Id: from_domain.py 1527 2017-02-15 18:36:36Z pe $
"""

from .base import EbasFileBase
class EbasFileFromDomain(EbasFileBase):
    """
    Fallback class to be used when ebas.domain is not available (e.g. EbasIO
    Module published for 3rd parties).
    """
    pass
